import json
import os

from pykechain.enums import PropertyType, Multiplicity
from pykechain.models import AttachmentProperty
from tests.classes import TestBetamax


class TestAttachment(TestBetamax):
    test_dict = {"a": 1, "b": 3}
    project_root = os.path.dirname(
        os.path.dirname(os.path.dirname(os.path.abspath(__file__)).replace("\\", "/"))
    )

    def setUp(self):
        super().setUp()
        self.property_name = "Plot Attachment"

        self.root_model = self.project.model(name="Product")
        self.part_model = self.root_model.add_model(
            name="__Testing attachments", multiplicity=Multiplicity.ONE_MANY
        )
        self.property_model = self.part_model.add_property(
            name=self.property_name, property_type=PropertyType.ATTACHMENT_VALUE
        )  # type: AttachmentProperty

        self.property = self.part_model.instance().property(
            name=self.property_name
        )  # type: AttachmentProperty

    def tearDown(self):
        self.part_model.delete()
        super().tearDown()

    def test_retrieve_attachment(self):
        data = ("data.json", json.dumps(self.test_dict), "application/json")

        self.property._upload(data)

        r = self.property.json_load()

        self.assertDictEqual(r, self.test_dict)

    def test_retrieve_value(self):
        self.assertIsNone(self.property.value)

        self.property.upload(self.project_root + "/requirements.txt")

        self.assertIsNotNone(self.property.value)

    def test_set_value_none(self):
        self.property.value = None
        self.assertFalse(self.property.has_value())

    def test_set_value_not_a_path(self):
        with self.assertRaises(FileNotFoundError):
            self.property.value = "String must be a filepath"

    def test_upload(self):
        self.property.upload(self.project_root + "/requirements.txt")

        requirements = self.project_root + "/requirements.txt"
        self.property.upload(requirements)

    # 1.11.1
    def test_clear_an_attachment_property(self):
        # setUp
        self.property.upload(self.project_root + "/requirements.txt")

        # testing
        self.assertTrue(self.property.has_value())

        self.property.clear()
        self.assertIsNone(self.property.value)
        self.assertIsNone(self.property._value)

    def test_retrieve_filename_from_value(self):
        # setUp
        self.property.upload(self.project_root + "/requirements.txt")

        # testing
        self.assertEqual(self.property.filename, "requirements.txt")

    def test_has_value_true(self):
        # setUp
        self.property.upload(self.project_root + "/requirements.txt")

        # testing
        self.assertTrue(self.property.has_value())

    def test_has_value_false(self):
        self.assertFalse(self.property.has_value())

    def test_add_with_properties(self):
        # setUp
        self.new_wheel = self.root_model.instance().add_with_properties(
            self.part_model,
            "new part",
            update_dict={self.property_name: self.project_root + "/requirements.txt"},
        )



from datetime import datetime
from typing import Dict, List, Optional, Union

import requests

from pykechain.enums import ContextGroup, ContextType
from pykechain.exceptions import APIError
from pykechain.models import BaseInScope
from pykechain.models.input_checks import (
    check_base,
    check_datetime,
    check_enum,
    check_list_of_base,
    check_list_of_text,
    check_text,
    check_type,
)
from pykechain.models.tags import TagsMixin
from pykechain.typing import ObjectID, ObjectIDs
from pykechain.utils import Empty, clean_empty_values, empty, parse_datetime


class Context(BaseInScope, TagsMixin):
    """
    A virtual object representing a KE-chain Context.

    .. versionadded:: 3.12
    """

    def __init__(self, json, **kwargs):
        """Construct a service from provided json data."""
        super().__init__(json, **kwargs)

        self.ref = json.get("ref")  # type: Text
        self.description = json.get("description", "")  # type:Text
        self.context_type = json.get("context_type")  # type: ContextType
        self.options = json.get("options", dict())  # type: Dict
        self.context_group = json.get("context_group", "")  # type: ContextGroup
        self._tags = json.get("tags")

        # associated activities
        self.activities = json.get("activities")  # type: ObjectIDs

        # LocationContext
        # these attributes are enabled when context_type == STATIC_LOCATION
        self.feature_collection = json.get("feature_collection", dict())  # type: dict

        # TimeperiodContext
        # these attributes are enabled when context_type == TIME_PERIOD
        self.start_date = parse_datetime(json.get("start_date"))  # type: Optional[datetime]
        self.due_date = parse_datetime(json.get("due_date"))  # type: Optional[datetime]

    def edit(
        self,
        name: Optional[Union[str, Empty]] = empty,
        description: Optional[Union[str, Empty]] = empty,
        tags: Optional[List[Union[str, Empty]]] = empty,
        context_group: Optional[Union[ContextGroup, Empty]] = empty,
        scope: Optional[Union["Scope", ObjectID]] = empty,
        options: Optional[dict] = empty,
        activities: Optional[Union[List["Activity"], ObjectIDs]] = empty,
        feature_collection: Optional[dict] = empty,
        start_date: Optional[datetime] = empty,
        due_date: Optional[datetime] = empty,
        **kwargs,
    ) -> "self":
        """
        Edit the Context.

        :param name: Name of the Context to be displayed to the end-user.
        :param scope: Scope object or Scope Id where the Context is active on.
        :param description: (optional) description of the Context
        :param activities: (optional) associated list of Activity or activity object ID
        :param tags: (optional) tags
        :param context_group: (optional) a context context_group of the choices of `ContextGroup`
        :param options: (optional) dictionary with options.
        :param feature_collection: (optional) dict with a geojson feature collection to store for a STATIC_LOCATION
        :param start_date: (optional) start datetime for a TIME_PERIOD context
        :param due_date: (optional) start datetime for a TIME_PERIOD context
        :return: a created Context Object
        :return: The updated Context object
        """
        from pykechain.models import Scope
        from pykechain.models import Activity

        update_dict = {
            "name": check_text(name, "name"),
            "description": check_text(description, "description"),
            "scope": check_base(scope, Scope, "scope"),
            "tags": check_list_of_text(tags, "tags"),
            "context_group": check_enum(context_group, ContextGroup, "context_group"),
            "activities": check_list_of_base(activities, Activity, "activities"),
            "options": check_type(options, dict, "options"),
            "feature_collection": check_type(feature_collection, dict, "feature_collection"),
            "start_date": check_datetime(start_date, "start_date"),
            "due_date": check_datetime(due_date, "due_date"),
        }

        if feature_collection is None:
            update_dict["feature_collection"] = {}

        if kwargs:  # pragma: no cover
            update_dict.update(kwargs)

        update_dict = clean_empty_values(update_dict=update_dict)

        url = self._client._build_url("context", context_id=self.id)

        response = self._client._request("PUT", url, json=update_dict)

        if response.status_code != requests.codes.ok:  # pragma: no cover
            raise APIError(f"Could not update Context: {self}", response=response)

        return self.refresh(json=response.json().get("results")[0])

    def link_activities(
        self, activities: Optional[List[Union["Activity", ObjectIDs]]] = empty, **kwargs
    ):
        """
        Link a context to one or more activities.

        :param activities: optional list of Activities or object Id's from activities.
        :returns: updated context objects
        """
        from pykechain.models import Activity

        update_dict = {
            "activities": check_list_of_base(activities, Activity, "activities"),
        }
        if kwargs:  # pragma: no cover
            update_dict.update(kwargs)

        url = self._client._build_url("context_link_activities", context_id=self.id)
        response = self._client._request("POST", url, json=update_dict)

        if response.status_code != requests.codes.ok:  # pragma: no cover
            raise APIError(f"Could not update Context: {self}", response=response)

        return self.refresh(json=response.json().get("results")[0])

    def unlink_activities(
        self, activities: Optional[List[Union["Activity", ObjectIDs]]] = empty, **kwargs
    ):
        """
        Unlink a context to one or more activities.

        :param activities: optional list of Activities or object Id's from activities.
        :returns: updated context objects
        """
        from pykechain.models import Activity

        update_dict = {
            "activities": check_list_of_base(activities, Activity, "activities"),
        }
        if kwargs:  # pragma: no cover
            update_dict.update(kwargs)

        url = self._client._build_url("context_unlink_activities", context_id=self.id)
        response = self._client._request("POST", url, json=update_dict)

        if response.status_code != requests.codes.ok:  # pragma: no cover
            raise APIError(f"Could not update Context: {self}", response=response)

        return self.refresh(json=response.json().get("results")[0])

    def delete(self):
        """Delete the Context."""
        return self._client.delete_context(self)



"""All validators and effect objects.

.. versionadded:: 2.2
"""

from .effects import (
    TextEffect,
    ErrorTextEffect,
    HelpTextEffect,
    VisualEffect,
    ValidVisualEffect,
    InvalidVisualEffect,
)
from .validators import (
    EmailValidator,
    FileExtensionValidator,
    FileSizeValidator,
    NumericRangeValidator,
    BooleanFieldValidator,
    RequiredFieldValidator,
    OddNumberValidator,
    EvenNumberValidator,
    RegexStringValidator,
    SingleReferenceValidator,
)
from .validators_base import PropertyValidator, ValidatorEffect  # noqa

__all__ = (
    "PropertyValidator",
    "ValidatorEffect",
    "TextEffect",
    "ErrorTextEffect",
    "HelpTextEffect",
    "VisualEffect",
    "ValidVisualEffect",
    "InvalidVisualEffect",
    "NumericRangeValidator",
    "BooleanFieldValidator",
    "RequiredFieldValidator",
    "OddNumberValidator",
    "EvenNumberValidator",
    "RegexStringValidator",
    "SingleReferenceValidator",
    "EmailValidator",
    "FileSizeValidator",
    "FileExtensionValidator",
)



from pykechain.enums import TeamRoles
from pykechain.exceptions import NotFoundError, MultipleFoundError, IllegalArgumentError
from pykechain.models import User
from pykechain.models.team import Team
from tests.classes import TestBetamax


class TestTeams(TestBetamax):
    def setUp(self):
        super().setUp()

        self.required_kwargs = dict(
            name="_test team", user=self.client.user("testuser")  # type: User
        )

        self.team = self.client.create_team(**self.required_kwargs)  # type: Team

    def tearDown(self):
        self.team.delete()
        super().tearDown()

    def test_create_team(self):
        self.assertIsInstance(self.team, Team)
        self.assertEqual(len(self.team.members()), 1)
        self.assertEqual(self.team.name, "_test team")

    def test_create_team_with_inputs(self):
        # setUp
        landing_page = f"#/scopes/{self.project.id}"
        new_team = self.client.create_team(
            description="This is the description",
            options=dict(
                landingPage=landing_page,
            ),
            is_hidden=True,
            **self.required_kwargs,
        )

        # testing
        self.assertEqual(new_team.description, "This is the description")
        self.assertDictEqual(new_team.options, {"landingPage": landing_page})
        self.assertTrue(new_team.is_hidden)

        # tearDown
        new_team.delete()

    def test_create_team_incorrect_inputs(self):
        with self.assertRaises(IllegalArgumentError):
            self.client.create_team(description=1, **self.required_kwargs)
        with self.assertRaises(IllegalArgumentError):
            self.client.create_team(options=f"#/scopes/{self.project.id}", **self.required_kwargs)
        with self.assertRaises(IllegalArgumentError):
            self.client.create_team(is_hidden="False", **self.required_kwargs)

    def test_retrieve_teams(self):
        self.assertTrue(self.client.teams())

    def test_retrieve_single_unknown_team(self):
        with self.assertRaises(NotFoundError):
            self.client.team("This is not a existing team name for sure")

    def test_retrieve_single_multiple_team_raises_error(self):
        with self.assertRaises(MultipleFoundError):
            self.client.team()

    def test_retrieve_single_team_with_known_teamname(self):
        self.assertTrue(self.client.team(name="_test team"))

    def test_retrieve_members(self):
        members = self.team.members()
        self.assertTrue(len(members) > 0)

    def test_retrieve_managers_members_owners(self):
        all_members = self.team.members()

        manager = self.team.members(role=TeamRoles.MANAGER)
        self.assertTrue(len(manager) >= 0)

        owner = self.team.members(role=TeamRoles.OWNER)
        self.assertTrue(len(owner) > 0)

        members = self.team.members(role=TeamRoles.MEMBER)
        self.assertTrue(len(owner) > 0)

        self.assertEqual(len(all_members), len(manager) + len(owner) + len(members))

    def test_retrieve_member_with_invalid_role(self):
        with self.assertRaisesRegex(IllegalArgumentError, "must be an option from enum"):
            self.team.members(role="FOOBARROLE")

    def test_add_and_remove_member(self):
        members = self.team.members()
        a_user = self.client.user(username="anotheruser")

        self.team.add_members([a_user.id], role=TeamRoles.MEMBER)

        self.assertIn(
            a_user.id, [member.get("pk") for member in self.team.members(role=TeamRoles.MEMBER)]
        )

        self.team.remove_members([a_user.id])

        self.assertListEqual(members, self.team.members())

    def test_add_scope_to_team(self):
        # setup
        old_team = self.project._json_data.get("team")

        self.project.edit(team=self.team.id)
        self.assertEqual(self.project.team, self.team)

        # check
        team_scopes = self.team.scopes()
        self.assertEqual([t.id for t in team_scopes], [self.project.id])

        # teardown
        self.project.edit(team=old_team and old_team.get("id") or None)

    def test_team_attributes(self):
        attributes = ["_client", "_json_data", "id", "name", "created_at", "updated_at", "ref"]

        obj = self.team
        for attribute in attributes:
            self.assertTrue(
                hasattr(obj, attribute),
                f"Could not find '{attribute}' in the object: '{obj.__dict__}'",
            )

    def test_team_edit(self):
        # setUp
        options = dict(landingPage=f"#/scopes/{self.project.id}")
        self.team.edit(
            name="renamed team",
            description="My team description",
            options=options,
            is_hidden=True,
        )

        # testing
        self.assertEqual(self.team.name, "renamed team")
        self.assertEqual(self.team.description, "My team description")
        self.assertDictEqual(self.team.options, options)
        self.assertTrue(self.team.is_hidden)

    # noinspection PyTypeChecker
    def test_team_edit_wrong_inputs(self):
        with self.assertRaises(IllegalArgumentError):
            self.team.edit(name=123)
        with self.assertRaises(IllegalArgumentError):
            self.team.edit(description=False)
        with self.assertRaises(IllegalArgumentError):
            self.team.edit(options="New scope")
        with self.assertRaises(IllegalArgumentError):
            self.team.edit(is_hidden=1)



import datetime

from pykechain.enums import (
    ScopeStatus,
    KEChainPages,
    ScopeMemberActions,
    ScopeRoles,
    ScopeCategory,
)
from pykechain.exceptions import NotFoundError, MultipleFoundError, IllegalArgumentError
from pykechain.models import Team, Scope
from pykechain.models.sidebar.sidebar_manager import SideBarManager
from pykechain.utils import is_url
from tests.classes import TestBetamax


class TestScopes(TestBetamax):
    """
    Test retrieval and scope attributes and methods.
    """

    def test_scope_attributes(self):
        attributes = [
            "_client",
            "_json_data",
            "id",
            "name",
            "created_at",
            "updated_at",
            "ref",
            "process",
            "workflow_root",
            "description",
            "status",
            "category",
            "tags",
            "start_date",
            "due_date",
        ]

        obj = self.project
        for attribute in attributes:
            with self.subTest(attribute):
                self.assertTrue(
                    hasattr(obj, attribute),
                    f"Could not find '{attribute}' in the object: '{obj.__dict__.keys()}'",
                )

    def test_retrieve_scopes(self):
        self.assertTrue(self.client.scopes())

    def test_retrieve_scope_with_refs(self):
        # setup
        scope_ref = "bike-project"
        scope_name = "Bike Project"
        scope = self.client.scope(ref=scope_ref)

        # testing
        self.assertIsInstance(scope, Scope)
        self.assertTrue(scope.name, scope_name)

    def test_retrieve_scope_with_kwargs(self):
        retrieve_scopes_with_kwargs = self.client.scopes(name__icontains="Bike")
        self.assertTrue(retrieve_scopes_with_kwargs)
        self.assertTrue(len(retrieve_scopes_with_kwargs) >= 1)

    def test_retrieve_single_unknown(self):
        with self.assertRaises(NotFoundError):
            self.client.scope("123lladadwd")

    def test_retrieve_single_multiple(self):
        with self.assertRaises(MultipleFoundError):
            self.client.scope()

    def test_team_property_of_scope(self):
        """Test for the property 'team' of a scope."""
        team = self.project.team

        self.assertIsNone(team)

    def test_side_bar(self):
        side_bar_manager = self.project.side_bar()

        self.assertIsInstance(side_bar_manager, SideBarManager)

    def test_root_properties(self):
        for value in [
            self.project.workflow_root_process,
            self.project.catalog_root_process,
            self.project.app_root_process,
            self.project.product_root_model,
            self.project.product_root_instance,
            self.project.catalog_root_model,
            self.project.catalog_root_instance,
        ]:
            with self.subTest(""):
                self.assertIsNotNone(value)


class TestScopeMembers(TestBetamax):
    """
    Tests user assignment to a scope.
    """

    DEFAULTS = {
        ScopeRoles.MANAGER: [
            "superuser",
            "testmanager",
        ],
        ScopeRoles.LEADMEMBER: [
            "superuser",
            "testlead",
        ],
        ScopeRoles.SUPERVISOR: ["supervisor"],
        ScopeRoles.MEMBER: [
            "testuser",
        ],
    }

    def _reset_members(self):
        """Reset the scope members by removing all members and adding the defaults"""
        [
            self.project.remove_member(member=member["username"])
            for member in self.project.members()
        ]

        for role in [
            ScopeRoles.MANAGER,
            ScopeRoles.LEADMEMBER,
            ScopeRoles.MEMBER,
            ScopeRoles.SUPERVISOR,
        ]:
            for member in self.DEFAULTS[role]:
                self.project._update_scope_project_team(ScopeMemberActions.ADD, role, member)

    def setUp(self):
        super().setUp()
        self._reset_members()

    def tearDown(self):
        self._reset_members()
        super().tearDown()

    def test_reset_members(self):
        """Test `_reset_members` method of the test class"""
        self.assertTrue(True)

    def test_members(self):
        scope = self.project

        # Helper function:
        def get_member_names(**kwargs):
            return [member["username"] for member in scope.members(**kwargs)]

        all_members = set(get_member_names())
        managers = set(get_member_names(is_manager=True))
        supervisors = set(get_member_names(is_supervisor=True))
        non_managers = set(get_member_names(is_manager=False))
        leads = set(get_member_names(is_leadmember=True))
        non_leads = set(get_member_names(is_leadmember=False))
        basic_members = set(get_member_names(is_manager=False, is_leadmember=False))

        self.assertEqual(5, len(all_members), msg=",".join(all_members))
        self.assertTrue(
            basic_members and leads and managers and supervisors,
            msg="Test scope must have a member of every type!",
        )

        all_roles = basic_members | leads | managers | supervisors
        remaining_non_leads = all_roles - leads
        remaining_non_managers = all_roles - managers

        self.assertSetEqual(
            all_roles, all_members, msg="The sum of all roles should equal all members."
        )
        self.assertSetEqual(remaining_non_leads, non_leads)
        self.assertSetEqual(remaining_non_managers, non_managers)

    def test_add_member(self):
        member_to_be_added = "anotheruser"
        self.project.add_member(member_to_be_added)
        self.project = self.client.scope(pk=self.project.id)
        project_members = self.project.members()

        # testing
        self.assertTrue(member_to_be_added in [member["username"] for member in project_members])

    def test_add_member_by_id(self):
        member_id = 4
        with self.assertRaises(IllegalArgumentError):
            # noinspection PyTypeChecker
            self.project.add_member(member_id)

    def test_add_non_existing_member(self):
        member_to_be_added = "Nobody"
        with self.assertRaises(NotFoundError):
            self.project.add_member(member_to_be_added)

    def test_remove_member(self):
        member_to_be_removed = "anotheruser"
        # setUp
        self.project.add_member(member_to_be_removed)
        # testing
        self.project.remove_member(member_to_be_removed)
        self.project = self.client.scope(pk=self.project.id)
        project_members = self.project.members()
        self.assertTrue(
            member_to_be_removed not in [member["username"] for member in project_members]
        )

    def test_add_manager(self):
        manager_to_be_added = "anotheruser"
        # testing
        self.project.add_manager(manager_to_be_added)
        self.project = self.client.scope(pk=self.project.id)
        project_managers = self.project.members(is_manager=True)
        self.assertTrue(
            manager_to_be_added in [manager["username"] for manager in project_managers]
        )

    def test_remove_manager(self):
        manager_to_be_removed = "anotheruser"
        # setUp
        self.project.add_manager(manager_to_be_removed)
        # testing
        self.project.remove_manager(manager_to_be_removed)
        self.project = self.client.scope(pk=self.project.id)
        project_managers = self.project.members(is_manager=True)
        self.assertTrue(
            manager_to_be_removed not in [manager["username"] for manager in project_managers]
        )
        project_managers = self.project.members(is_manager=False)
        self.assertTrue(
            manager_to_be_removed in [manager["username"] for manager in project_managers]
        )

    def test_add_leadmember(self):
        leadmember_to_be_added = "anotheruser"
        # testing
        self.project.add_leadmember(leadmember_to_be_added)
        self.project = self.client.scope(pk=self.project.id)
        project_leadmembers = self.project.members(is_leadmember=True)
        self.assertTrue(
            leadmember_to_be_added
            in [leadmember["username"] for leadmember in project_leadmembers]
        )

    def test_remove_leadmember(self):
        leadmember_to_be_removed = "anotheruser"
        # setUp
        self.project.add_leadmember(leadmember_to_be_removed)
        # testing
        self.project.remove_leadmember(leadmember_to_be_removed)
        self.project = self.client.scope(pk=self.project.id)
        project_leadmembers = self.project.members(is_leadmember=True)
        self.assertTrue(
            leadmember_to_be_removed
            not in [leadmember["username"] for leadmember in project_leadmembers]
        )
        project_leadmembers = self.project.members(is_leadmember=False)
        self.assertTrue(
            leadmember_to_be_removed
            in [leadmember["username"] for leadmember in project_leadmembers]
        )

    def test_add_supervisor(self):
        supervisor_to_be_added = "anotheruser"
        # testing
        self.project.add_supervisor(supervisor_to_be_added)
        self.project = self.client.scope(pk=self.project.id)
        project_supervisors = self.project.members(is_supervisor=True)
        self.assertTrue(
            supervisor_to_be_added
            in [supervisor["username"] for supervisor in project_supervisors]
        )

    def test_remove_supervisor(self):
        supervisor_to_be_removed = "anotheruser"
        # setUp
        self.project.add_supervisor(supervisor_to_be_removed)
        # testing
        self.project.remove_supervisor(supervisor_to_be_removed)
        self.project = self.client.scope(pk=self.project.id)
        project_supervisors = self.project.members(is_supervisor=True)
        self.assertTrue(
            supervisor_to_be_removed
            not in [supervisor["username"] for supervisor in project_supervisors]
        )


class TestScopeEdit(TestBetamax):
    """
    Test modifications to a scope.
    """

    def setUp(self):
        super().setUp()
        testing_project = self.project
        self.project = None
        self.scope = testing_project.clone()

    def tearDown(self):
        if self.scope:
            self.scope.delete()
        super().tearDown()

    # 2.2.0+
    # noinspection PyTypeChecker
    def test_edit_scope(self):
        # setUp
        new_scope_name = "Pykechain testing (bike project)"
        new_scope_description = (
            "Project used to build a Bike. Part-time job is to also test Pykechain."
        )
        new_start_date = datetime.datetime(2018, 12, 5, tzinfo=None)
        new_due_date = datetime.datetime(2018, 12, 8, tzinfo=None)
        new_tags = ["tag_one", "tag_two"]

        self.scope.edit(
            name=new_scope_name,
            description=new_scope_description,
            tags=new_tags,
            start_date=new_start_date,
            due_date=new_due_date,
            status=ScopeStatus.CLOSED,
            category=ScopeCategory.TEMPLATE_SCOPE,
        )

        retrieved_project = self.client.scope(id=self.scope.id, status=ScopeStatus.CLOSED)

        # testing
        self.assertEqual(new_scope_name, retrieved_project.name)
        self.assertEqual(new_scope_description, retrieved_project.description)
        self.assertIn(
            retrieved_project._json_data["start_date"],
            ("2018-12-05T00:00:00Z", "2018-12-05T00:00:00+00:00"),
        )
        self.assertIn(
            retrieved_project._json_data["due_date"],
            ("2018-12-08T00:00:00Z", "2018-12-08T00:00:00+00:00"),
        )
        self.assertEqual(ScopeStatus.CLOSED, retrieved_project.status)
        self.assertEqual(ScopeCategory.TEMPLATE_SCOPE, retrieved_project.category)
        self.assertEqual(new_tags, retrieved_project.tags)

        with self.assertRaises(IllegalArgumentError):
            self.scope.edit(name=True)

        with self.assertRaises(IllegalArgumentError):
            self.scope.edit(description=new_start_date)

        with self.assertRaises(IllegalArgumentError):
            self.scope.edit(start_date="Yes, man!")

        with self.assertRaises(IllegalArgumentError):
            self.scope.edit(due_date="Wrong")

        with self.assertRaises(IllegalArgumentError):
            self.scope.edit(status="ILLEGAL_STATUS_WILL_CAUSE_ERR")

        with self.assertRaises(IllegalArgumentError):
            self.scope.edit(category="Ice scoop")

        with self.assertRaises(IllegalArgumentError):
            self.scope.edit(tags="tags must be a list of strings")

    # test added due to #847 - providing no inputs overwrites values
    def test_edit_scope_clearing_values(self):
        # setUp
        initial_name = "Pykechain testing (bike project)"
        initial_description = (
            "Project used to build a Bike. Part-time job is to also test Pykechain."
        )
        initial_start_date = datetime.datetime(2018, 12, 5, tzinfo=None)
        initial_due_date = datetime.datetime(2018, 12, 8, tzinfo=None)
        initial_tags = ["tag_one", "tag_two"]
        team_one, team_two = self.client.teams()[:2]

        self.scope.edit(
            name=initial_name,
            description=initial_description,
            tags=initial_tags,
            start_date=initial_start_date,
            due_date=initial_due_date,
            team=team_two,
        )

        # Edit without mentioning values, everything should stay the same
        new_name = "Just Pykechain testing (bike project)"
        self.scope.edit(name=new_name)

        # testing
        self.assertEqual(self.scope.name, new_name)
        self.assertEqual(self.scope.description, initial_description)
        self.assertEqual(
            self.scope.start_date.strftime("%Y/%m/%d, %H:%M:%S"),
            initial_start_date.strftime("%Y/%m/%d, %H:%M:%S"),
        )
        self.assertEqual(
            self.scope.due_date.strftime("%Y/%m/%d, %H:%M:%S"),
            initial_due_date.strftime("%Y/%m/%d, %H:%M:%S"),
        )
        self.assertEqual(self.scope.tags, initial_tags)

        # Edit with clearing the values, name and status cannot be cleared
        self.scope.edit(
            name=None, description=None, tags=None, start_date=None, due_date=None, status=None
        )
        self.scope.refresh()
        self.assertEqual(self.scope.name, new_name)
        self.assertEqual(self.scope.description, "")
        self.assertEqual(self.scope.start_date, None)
        self.assertEqual(self.scope.due_date, None)
        self.assertEqual(self.scope.tags, list())

    # v2.3.3
    def test_edit_scope_team(self):
        """Test capabilities of changing team scope"""

        team_one, team_two = self.client.teams()[:2]

        self.scope.edit(team=team_one)
        self.assertEqual(self.scope.team, team_one)

        self.scope.edit(team=team_two)
        self.assertEqual(self.scope.team, team_two)

    def test_set_landing_page(self):
        tasks = [
            self.scope.activities()[0],
            KEChainPages.DATA_MODEL,
        ]

        for task in tasks:
            with self.subTest(msg=task):
                self.scope.set_landing_page(activity=task)

    def test_get_landing_page(self):
        landing_page = self.scope.get_landing_page_url()

        self.assertIsNone(landing_page)

        self.scope.set_landing_page(activity=self.scope.activities()[0])
        landing_page = self.scope.get_landing_page_url()

        self.assertIsInstance(landing_page, str)
        self.assertTrue(is_url(self.client.api_root + landing_page))



import os
from datetime import datetime

import requests
from typing import Dict, Optional, Union

from pykechain.enums import (
    ServiceScriptUser,
    ServiceExecutionStatus,
    ServiceType,
    ServiceEnvironmentVersion,
)
from pykechain.exceptions import APIError
from pykechain.models.base import Base, BaseInScope
from pykechain.models.input_checks import check_text, check_enum, check_type
from pykechain.utils import parse_datetime, Empty, clean_empty_values, empty


class Service(BaseInScope):
    """
    A virtual object representing a KE-chain Service.

    .. versionadded:: 1.13

    :ivar id: id of the service
    :type id: uuid
    :ivar name: name of the service
    :type name: str
    :ivar description: description of the service
    :type description: str
    :ivar version: version number of the service, as provided by uploaded
    :type version: str
    :ivar type: type of the service. One of the :class:`ServiceType`
    :type type: str
    :ivar filename: filename of the service
    :type filename: str
    :ivar environment: environment in which the service will execute. One of :class:`ServiceEnvironmentVersion`
    :type environment: str
    :ivar updated_at: datetime in UTC timezone when the Service was last updated
    :type updated_at: datetime

    .. versionadded:: 3.0

    :ivar trusted: Trusted flag. If the kecpkg is trusted.
    :ivar run_as: User to run the script as. One of :class:`ServiceScriptUser`.
    :ivar verified_on: Date when the kecpkg was verified by KE-chain (if verification pipeline is enabled)
    :ivar verification_results: Results of the verification (if verification pipeline is enabled)
    """

    def __init__(self, json, **kwargs):
        """Construct a service from provided json data."""
        super().__init__(json, **kwargs)
        del self.created_at

        self.description = json.get("description", "")
        self.version = json.get("script_version", "")
        self.filename = json.get("script_file_name")
        self.type = json.get("script_type")
        self.environment = json.get("env_version")

        # for SIM3 version
        self.trusted: bool = json.get("trusted")
        self.run_as: str = json.get("run_as")
        self.verified_on: Optional[datetime] = parse_datetime(json.get("verified_on"))
        self.verification_results: Dict = json.get("verification_results")

    def __repr__(self):  # pragma: no cover
        return f"<pyke Service '{self.name}' id {self.id[-8:]}>"

    def execute(self, interactive: Optional[bool] = False, **kwargs) -> "ServiceExecution":
        """
        Execute the service.

        For interactive (notebook) service execution, set interactive to True, defaults to False.

        .. versionadded:: 1.13

        :param interactive: (optional) True if the notebook service should execute in interactive mode.
        :type interactive: bool or None
        :return: ServiceExecution when successful.
        :raises APIError: when unable to execute
        """
        request_params = dict(
            interactive=check_type(interactive, bool, "interactive"), format="json", **kwargs
        )

        url = self._client._build_url("service_execute", service_id=self.id)
        response = self._client._request("GET", url, params=request_params)

        if response.status_code == requests.codes.conflict:  # pragma: no cover
            raise APIError(
                f"Conflict: Could not execute Service {self} as it is already running.",
                response=response,
            )
        elif response.status_code != requests.codes.accepted:  # pragma: no cover
            raise APIError(f"Could not execute Service {self}", response=response)

        data = response.json()
        return ServiceExecution(json=data.get("results")[0], client=self._client)

    def edit(
        self,
        name: Optional[Union[str, Empty]] = empty,
        description: Optional[Union[str, Empty]] = empty,
        version: Optional[Union[str, Empty]] = empty,
        type: Optional[Union[ServiceType, Empty]] = empty,
        environment_version: Optional[Union[ServiceEnvironmentVersion, Empty]] = empty,
        run_as: Optional[Union[ServiceScriptUser, Empty]] = empty,
        trusted: Optional[Union[bool, Empty]] = empty,
        **kwargs,
    ) -> None:
        """
        Edit Service details.

        Setting an input to None will clear out the value (exception being name).

        .. versionadded:: 1.13

        :param name: (optional) name of the service to change. Cannot be cleared.
        :type name: basestring or None or Empty
        :param description: (optional) description of the service. Can be cleared.
        :type description: basestring or None or Empty
        :param version: (optional) version number of the service. Can be cleared.
        :type version: basestring or None or Empty
        :param type: (optional) script type (Python or Notebook). Cannot be cleared.
        :type type: ServiceType or None or Empty
        :param environment_version: (optional) environment version of the service. Cannot be cleared.
        :type environment_version: ServiceEnvironmentVersion or None or Empty
        :param run_as: (optional) user to run the service as. Defaults to kenode user (bound to scope).
                    Cannot be cleared.
        :type run_as: ServiceScriptUser or None or Empty
        :param trusted: (optional) flag whether the service is trusted, default if False. Cannot be cleared.
        :type trusted: bool or None or Empty
        :raises IllegalArgumentError: when you provide an illegal argument.
        :raises APIError: if the service could not be updated.

        Example
        -------
        >>> service.edit(name='Car service',version='203')

        Not mentioning an input parameter in the function will leave it unchanged. Setting a parameter as None will
        clear its value (where that is possible). The example below will clear the description and edit the name.

        >>> service.edit(name="Plane service",description=None)

        """
        update_dict = {
            "id": self.id,
            "name": check_text(name, "name") or self.name,
            "description": check_text(description, "description") or "",
            "trusted": check_type(trusted, bool, "trusted") or self.trusted,
            "script_type": check_enum(type, ServiceType, "type") or self.type,
            "env_version": check_enum(
                environment_version, ServiceEnvironmentVersion, "environment version"
            )
            or self.environment,
            "run_as": check_enum(run_as, ServiceScriptUser, "run_as") or self.run_as,
            "script_version": check_text(version, "version") or "",
        }

        if kwargs:  # pragma: no cover
            update_dict.update(**kwargs)

        update_dict = clean_empty_values(update_dict=update_dict)

        response = self._client._request(
            "PUT", self._client._build_url("service", service_id=self.id), json=update_dict
        )

        if response.status_code != requests.codes.ok:  # pragma: no cover
            raise APIError(f"Could not update Service {self}", response=response)

        self.refresh(json=response.json()["results"][0])

    def delete(self) -> None:
        """Delete this service.

        :raises APIError: if delete was not successful.
        """
        response = self._client._request(
            "DELETE", self._client._build_url("service", service_id=self.id)
        )

        if response.status_code != requests.codes.no_content:  # pragma: no cover
            raise APIError(f"Could not delete Service {self}", response=response)

    def upload(self, pkg_path):
        """
        Upload a python script (or kecpkg) to the service.

        .. versionadded:: 1.13

        :param pkg_path: path to the python script or kecpkg to upload.
        :type pkg_path: basestring
        :raises APIError: if the python package could not be uploaded.
        :raises OSError: if the python package could not be located on disk.
        """
        if os.path.exists(pkg_path):
            self._upload(pkg_path=pkg_path)
        else:
            raise OSError(f"Could not locate python package to upload in '{pkg_path}'")

    def _upload(self, pkg_path):
        url = self._client._build_url("service_upload", service_id=self.id)

        with open(pkg_path, "rb") as pkg:
            response = self._client._request(
                "POST", url, files={"attachment": (os.path.basename(pkg_path), pkg)}
            )

        if response.status_code != requests.codes.accepted:  # pragma: no cover
            raise APIError(
                f"Could not upload script file (or kecpkg) to Service {self}", response=response
            )

        self.refresh(json=response.json()["results"][0])

    def save_as(self, target_dir=None):
        """
        Save the kecpkg service script to an (optional) target dir.

        Retains the filename of the service as known in KE-chain.

        .. versionadded:: 1.13

        :param target_dir: (optional) target dir. If not provided will save to current working directory.
        :type target_dir: basestring or None
        :raises APIError: if unable to download the service.
        :raises OSError: if unable to save the service kecpkg file to disk.
        """
        full_path = os.path.join(target_dir or os.getcwd(), self.filename)

        url = self._client._build_url("service_download", service_id=self.id)
        response = self._client._request("GET", url)
        if response.status_code != requests.codes.ok:  # pragma: no cover
            raise APIError(
                f"Could not download script file from Service {self}", response=response
            )

        with open(full_path, "w+b") as f:
            for chunk in response:
                f.write(chunk)

    def get_executions(self, **kwargs):
        """
        Retrieve the executions related to the current service.

        .. versionadded:: 1.13

        :param kwargs: (optional) additional search keyword arguments to limit the search even further.
        :type kwargs: dict
        :return: list of ServiceExecutions associated to the current service.
        """
        return self._client.service_executions(service=self.id, scope=self.scope_id, **kwargs)


class ServiceExecution(Base):
    """
    A virtual object representing a KE-chain Service Execution.

    .. versionadded:: 1.13

    :ivar id: id of the service execution
    :type id: uuid
    :ivar name: name of the service to which the execution is associated
    :type name: str
    :ivar status: status of the service. One of :class:`ServiceExecutionStatus`
    :type status: str
    :ivar service: the :class:`Service` object associated to this service execution
    :type service: :class:`Service`
    :ivar service_id: the uuid of the associated Service object
    :type service_id: uuid
    :ivar user: (optional) username of the user that executed the service
    :type user: str or None
    :ivar activity_id: (optional) the uuid of the activity where the service was executed from
    :type activity_id: uuid or None
    """

    def __init__(self, json, **kwargs):
        """Construct a scope from provided json data."""
        super().__init__(json, **kwargs)
        del self.created_at
        del self.updated_at

        self.name = json.get("service_name")
        self.service_id = json.get("service")
        self.status: ServiceExecutionStatus = json.get("status", "")

        self.user = json.get("username")
        if json.get("activity") is not None:
            self.activity_id: Optional[str] = json["activity"].get("id")
        else:
            self.activity_id = None

        self.started_at: Optional[datetime] = parse_datetime(json.get("started_at"))
        self.finished_at: Optional[datetime] = parse_datetime(json.get("finished_at"))

        self._service: Optional[Service] = None

    def __repr__(self):  # pragma: no cover
        return f"<pyke ServiceExecution '{self.name}' id {self.id[-8:]}>"

    @property
    def service(self) -> Service:
        """Retrieve the `Service` object to which this execution is associated."""
        if not self._service:
            self._service = self._client.service(id=self.service_id)
        return self._service

    def terminate(self):
        """
        Terminate the Service execution.

        .. versionadded:: 1.13

        :return: None if the termination request was successful
        :raises APIError: When the service execution could not be terminated.
        """
        url = self._client._build_url("service_execution_terminate", service_execution_id=self.id)
        response = self._client._request("GET", url, params=dict(format="json"))

        if response.status_code != requests.codes.accepted:  # pragma: no cover
            raise APIError(f"Could not terminate Service {self}", response=response)

    def get_log(self, target_dir=None, log_filename="log.txt"):
        """
        Retrieve the log of the service execution.

        .. versionadded:: 1.13

        :param target_dir: (optional) directory path name where the store the log.txt to.
        :type target_dir: basestring or None
        :param log_filename: (optional) log filename to write the log to, defaults to `log.txt`.
        :type log_filename: basestring or None
        :raises APIError: if the logfile could not be found.
        :raises OSError: if the file could not be written.
        """
        full_path = os.path.join(target_dir or os.getcwd(), log_filename)

        url = self._client._build_url("service_execution_log", service_execution_id=self.id)
        response = self._client._request("GET", url)
        if response.status_code != requests.codes.ok:  # pragma: no cover
            raise APIError(
                f"Could not download execution log of Service {self}", response=response
            )

        with open(full_path, "w+b") as f:
            for chunk in response:
                f.write(chunk)

    def get_notebook_url(self):
        """
        Get the url of the notebook, if the notebook is executed in interactive mode.

        .. versionadded:: 1.13

        :return: full url to the interactive running notebook as `basestring`
        :raises APIError: when the url cannot be retrieved.
        """
        url = self._client._build_url(
            "service_execution_notebook_url", service_execution_id=self.id
        )
        response = self._client._request("GET", url, params=dict(format="json"))

        if response.status_code != requests.codes.ok:
            raise APIError(f"Could not retrieve notebook url of Service {self}", response=response)

        data = response.json()
        url = data.get("results")[0].get("url")
        return url



from pykechain.enums import PropertyType
from pykechain.exceptions import IllegalArgumentError
from pykechain.models import SelectListProperty, Property
from pykechain.models.property_selectlist import MultiSelectListProperty
from tests.classes import TestBetamax


class SelectListBaseTests:
    """This wrapping class is excluded from the test runs, but can be inherited from."""

    class Tests(TestBetamax):
        """Tests for the select list properties."""

        PROPERTY_TYPE = None
        OPTIONS = ["1", "3.14", "text"]

        def setUp(self):
            super().setUp()

            self.select_model = self.project.model("Bike").add_property(
                name=self.PROPERTY_TYPE,
                property_type=self.PROPERTY_TYPE,
                options={"value_choices": ["1", "3.14", "text"]},
            )
            self.select = self.project.part("Bike").property(self.PROPERTY_TYPE)

        def tearDown(self):
            self.select_model.delete()
            super().tearDown()

        def test_get_options_list(self):
            self.assertTrue(hasattr(self.select_model, "options"))
            self.assertIsInstance(self.select_model.options, list)
            for item in self.select_model.options:
                self.assertTrue(type(item), str)

        def test_set_options_list(self):
            # setUp
            current_options_list = [1, 3.14, "text"]
            new_options_list = ["some", "new", 4, "options"]
            self.select_model.options = new_options_list

            # testing
            self.assertListEqual(self.select_model.options, list(map(str, new_options_list)))

            # teardown
            self.select_model.options = current_options_list

        def test_illegal_options_are_not_set(self):
            """Test for several illegal lists to be set"""
            with self.assertRaises(IllegalArgumentError):
                # not a list
                self.select_model.options = 1

            with self.assertRaises(IllegalArgumentError):
                self.select_model.options = None

            with self.assertRaises(IllegalArgumentError):
                # set
                self.select_model.options = {1, 2, 3}

            with self.assertRaises(IllegalArgumentError):
                # dict
                self.select_model.options = {"a": 1, "b": 2, "c": 3}

            with self.assertRaises(IllegalArgumentError):
                # tuple
                self.select_model.options = (1,)

        def test_fail_to_set_options_on_instance(self):
            """Test settings options on a property instance, only models are allowed options to be set"""
            self.assertTrue(hasattr(self.select, "options"))

            with self.assertRaises(IllegalArgumentError):
                self.select.options = [1, 3.14]

        def test_integrity_options_dict(self):
            # setUp
            original_options_dict = dict(self.select_model._options)
            self.select_model.options = list(self.select_model.options)

            # testing
            testing_options_dict = self.select_model._options
            self.assertEqual(original_options_dict, testing_options_dict)


class TestPropertySelectListProperty(SelectListBaseTests.Tests):
    PROPERTY_TYPE = PropertyType.SINGLE_SELECT_VALUE

    def test_value(self):
        self.select.value = self.OPTIONS[0]
        self.assertIsInstance(self.select, SelectListProperty)

    def test_pend_value(self):
        self.select.value = None

        Property.set_bulk_update(True)

        self.select.value = self.OPTIONS[1]

        live_property = self.project.property(id=self.select.id)
        self.assertFalse(live_property.has_value())

        Property.update_values(client=self.client)

        live_property.refresh()
        self.assertTrue(live_property.has_value())

    # in 1.16
    def test_set_value_in_options(self):
        # setUp
        current_options_list = self.select_model.options
        first_option = current_options_list[0]
        self.select.value = first_option

        # testing
        option_that_is_set = self.select.value
        self.assertEqual(first_option, option_that_is_set)

    # in 1.15
    def test_set_value_not_in_options_raises_error(self):
        with self.assertRaises(IllegalArgumentError):
            self.select.value = (
                "Some illegal value that is not inside the list of options for sure"
            )


class TestPropertyMultiSelectListProperty(SelectListBaseTests.Tests):
    PROPERTY_TYPE = PropertyType.MULTI_SELECT_VALUE

    def test_value(self):
        self.select.value = self.OPTIONS[0:1]
        self.assertIsInstance(self.select, MultiSelectListProperty)

    # in 1.16
    def test_set_value_in_options(self):
        # setUp
        current_options_list = self.select_model.options
        selection = current_options_list[1:2]
        self.select.value = selection

        # testing
        option_that_is_set = self.select.value
        self.assertEqual(selection, option_that_is_set)

    def test_set_value_not_in_options_raises_error(self):
        with self.assertRaises(IllegalArgumentError):
            self.select.value = [
                "Some illegal value that is not inside the list of options for sure"
            ]



from abc import abstractmethod
from typing import Iterable, Optional, List

from pykechain.exceptions import IllegalArgumentError
from pykechain.models.input_checks import check_list_of_text, check_type


class TagsMixin:
    """
    Mix-in class to enable the use of tags on an object.

    :ivar tags: list of tags
    :type tags: list
    """

    _tags: List[str] = list()

    @abstractmethod
    def edit(self, tags: Optional[Iterable[str]] = None, *args, **kwargs) -> None:
        """Edit the list of Tags."""
        pass  # pragma: no cover

    @property
    def tags(self) -> List[str]:
        """
        Get a list of tags, with each tag being a string.

        :return: list of tags
        :rtype list
        """
        return list(self._tags) if self._tags is not None else list()

    @tags.setter
    def tags(self, new_tags: Iterable[str]) -> None:
        unique_tags = check_list_of_text(new_tags, "tags", True)
        self.edit(tags=unique_tags)
        self._tags = unique_tags

    def remove_tag(self, tag: str) -> None:
        """
        Remove a tag from the existing tags.

        :param tag: Tag to be removed from existing tags.
        :type tag: str
        :return: None
        """
        check_type(tag, str, "tag")

        if tag not in self.tags:
            raise IllegalArgumentError(
                "Tag '{}' is not among the existing tags. Existing tags: '{}'.".format(
                    tag, "', '".join(self.tags)
                )
            )

        remaining_tags = self.tags
        remaining_tags.remove(tag)
        self.tags = remaining_tags

    def add_tag(self, tag: str) -> None:
        """
        Append a tag to the existing tags.

        :param tag: Tag to be added to the existing tags.
        :type tag: str
        :return: None
        """
        check_type(tag, str, "tag")
        updated_tags = self.tags
        updated_tags.append(tag)
        self.tags = updated_tags

    def has_tag(self, tag: str) -> bool:
        """
        Check whether a tag is used.

        :param tag: Tag to be searched for.
        :return: boolean
        :rtype bool
        """
        check_type(tag, str, "tag")
        return tag in self.tags



from abc import abstractmethod
from copy import deepcopy
from typing import Any, Iterable

from jsonschema import validate

from pykechain.enums import Category
from pykechain.exceptions import APIError, IllegalArgumentError
from pykechain.models.property import Property
from pykechain.models.validators.validator_schemas import options_json_schema


class _SelectListProperty(Property):
    """A select list property that needs to update its options."""

    def __init__(self, json, **kwargs):
        """Construct a Property from a json object."""
        super().__init__(json, **kwargs)

        self._value_choices = json.get("value_options").get("value_choices")

    @property
    def value(self):
        """Retrieve the data value of a property.

        Setting this value will immediately update the property in KE-chain.
        The value should be in the list of options.

        :raises APIError: when unable to set the value or the value is not in the list of options
        """
        return deepcopy(self._value)

    @value.setter
    def value(self, value: Any):
        if value is not None:
            self._check_new_value(value=value)

        if self.use_bulk_update:
            self._pend_update(dict(value=value))
            self._value = value
        else:
            self._put_value(value)

    @abstractmethod
    def _check_new_value(self, value: Any):
        """
        Abstract method to validate the new value.

        :param value: any new value
        :return: sanitized value
        :raises APIError: when unable to set the value or the value is not in the list of options
        """
        pass  # pragma: no cover

    @property
    def options(self):
        """
        Get or set the options of the select list property if the property is a property model.

        Will raise an exception if the property is not a model. Will also raise an exception if the list is of
        incorrect form and contains non string, floats or integers.

        :param options_list: list of options to set, using only strings, integers and floats in the list.
        :type options_list: list
        :raises APIError: When unable to set the list of options
        :raises IllegalArgumentError: When the options_list is not in the right form or type, or the property is not
                                      a property model (:class:`Property` with category MODEL)

        Example
        -------
        >>> a_model = project.model('Model')
        >>> select_list_property = a_model.property('a_select_list_property')

        Get the list of options on the property
        >>> select_list_property.options
        ['1', '2', '3', '4']

        Update the list of options
        >>> select_list_property.options = [1,3,5,"wow"]

        List of options of this property to select from.

        """
        return deepcopy(self._value_choices)

    @options.setter
    def options(self, options_list):
        if not isinstance(options_list, list):
            raise IllegalArgumentError("The list of options should be a list")
        if self._json_data.get("category") != Category.MODEL:
            raise IllegalArgumentError(
                "We can only update the options list of the model of this property. The "
                "model of this property has id '{}'".format(self._json_data.get("model"))
            )

        # stringify the options list
        options_list = list(map(str, options_list))

        # check if the options are not already set and indeed different
        if not self._value_choices or set(options_list) != set(self._value_choices):
            self._put_value_options(options_list=options_list)
            self._value_choices = options_list

    def _put_value_options(self, options_list):
        new_options = (
            self._options.copy()
        )  # make a full copy of the dict not to only link it and update dict in place
        new_options.update({"value_choices": options_list})

        self._put_options(new_options)

    def _save_representations(self, representation_options):
        new_options = (
            self._options.copy()
        )  # make a full copy of the dict not to only link it and update dict in place
        new_options.update({"representations": representation_options})

        self._put_options(new_options)

    def _put_options(self, new_options):
        """Save the options to KE-chain.

        Makes a single API call.

        :param new_options: list of options to set.
        :raises APIError: when unable to update the options
        """
        validate(new_options, options_json_schema)

        url = self._client._build_url("property", property_id=self.id)
        response = self._client._request("PUT", url, json={"value_options": new_options})

        if response.status_code != 200:  # pragma: no cover
            raise APIError(f"Could not update options of Property {self}", response=response)
        else:
            self._options = new_options  # save the new options as the options


class SelectListProperty(_SelectListProperty):
    """A virtual object representing a KE-chain single-select list property."""

    def _check_new_value(self, value: Any):
        if value not in self.options:
            raise IllegalArgumentError(
                'The new value "{}" of the Property should be in the list of options:\n{}'.format(
                    value, self.options
                )
            )


class MultiSelectListProperty(_SelectListProperty):
    """A virtual object representing a KE-chain multi-select list property."""

    def _check_new_value(self, value: Iterable[Any]):
        if not isinstance(value, (list, tuple)):
            raise IllegalArgumentError(
                f'The new values must be provided as a list or tuple, "{value}" is not.'
            )

        if not all(v in self.options for v in value):
            raise IllegalArgumentError(
                'The new values "{}" of the Property should be in the list of options:\n{}'.format(
                    value, self.options
                )
            )



predefined_mimes = {
    "image/*": [
        "image/png",
        "image/gif",
        "image/jpeg",
        "image/webp",
        "image/tiff",
        "image/svg+xml",
        "image/bmp",
    ],
    ".xls": [
        "application/vnd.ms-excel",
        "application/x-excel",
        "application/x-msexcel",
        "application/vnd.openxmlformats-officedocument.spreadsheetml.sheet",
    ],
    ".xlsx": [
        "application/vnd.ms-excel",
        "application/x-excel",
        "application/x-msexcel",
        "application/vnd.openxmlformats-officedocument.spreadsheetml.sheet",
    ],
    "video/*": [
        "video/x-flv",
        "video/mp4",
        "application/x-mpegURL",
        "video/MP2T",
        "video/3gpp",
        "video/quicktime",
        "video/x-msvideo",
        "video/x-ms-wmv",
        "video/mpeg",
        "video/mp1s",
        "video/ogg",
        "video/webm",
    ],
    "model/*": [
        "model/3mf",
        "model/example",
        "model/gltf-binary",
        "model/gltf+json",
        "model/iges",
        "model/stl",
        "model/vnd.collada+xml",
        "model/vnd.dwf",
        "model/vnd.flatland.3dml",
        "model/vnd.gdl",
        "model/vnd.gs-gdl",
        "model/vnd.gtw",
        "model/vnd.moml+xml",
        "model/vnd.mts",
        "model/vnd.opengex",
        "model/vnd.usdz+zip",
        "model/vnd.valve.source.compiled-map",
        "model/vnd.vtu",
        "model/x3d-vrml",
        "model/x3d+fastinfoset",
        "model/x3d+xml",
    ],
}



import datetime
import urllib
import warnings
from unittest import TestCase

import pytz

from pykechain.enums import (
    FilterType,
    ScopeStatus,
    Multiplicity,
    PropertyType,
    ActivityType,
    ActivityRootNames,
)
from pykechain.exceptions import IllegalArgumentError
from pykechain.models import PropertyValueFilter
from pykechain.models.value_filter import ScopeFilter
from pykechain.models.widgets.enums import MetaWidget
from tests.classes import TestBetamax

TIMESTAMP = datetime.datetime(year=2020, month=1, day=1, tzinfo=pytz.UTC)
TIMESTAMP_2 = datetime.datetime(year=2020, month=5, day=1, tzinfo=pytz.UTC)


class TestPropertyValueFilter(TestBetamax):
    def setUp(self):
        super().setUp()

        self.bike = self.project.product_root_model("Bike")

        self.filter = PropertyValueFilter(
            property_model=self.bike.property(name="Gears"),
            value=15,
            filter_type=FilterType.GREATER_THAN_EQUAL,
        )

        self.new_part = self.bike.add_model(
            name="__TEST_PART__", multiplicity=Multiplicity.ZERO_MANY
        )
        self.new_test_property = self.new_part.add_property(
            name="__test_propERTY__", property_type=PropertyType.TEXT_VALUE
        )

    def tearDown(self):
        self.new_part.delete()
        super().tearDown()

    # noinspection PyTypeChecker
    def test_creation(self):
        with self.assertRaises(IllegalArgumentError):
            PropertyValueFilter(property_model=self.bike, value=15, filter_type=FilterType.EXACT)

        with self.assertRaises(IllegalArgumentError):
            PropertyValueFilter(
                property_model=self.bike.property(name="Gears"),
                value=15,
                filter_type="this one plz",
            )

    def test_repr(self):
        representation = self.filter.__repr__()

        self.assertIsInstance(representation, str)

    def test_format(self):
        string = self.filter.format()

        self.assertIsInstance(string, str)
        self.assertIn(":", string)

    def test_validate(self):
        self.filter.validate(part_model=self.bike)

        with self.assertRaises(IllegalArgumentError):
            self.filter.validate(part_model="Not a part model")

        bike_instance = self.bike.instance()
        property_instance = bike_instance.property(name="Gears")
        invalid_filter = PropertyValueFilter(
            property_model=property_instance,
            value=15,
            filter_type=FilterType.GREATER_THAN_EQUAL,
        )

        with self.assertRaises(IllegalArgumentError, msg="Property instance is not allowed"):
            invalid_filter.validate(part_model=bike_instance)

        with self.assertRaises(
            IllegalArgumentError, msg="Property instance should not match the bike model"
        ):
            invalid_filter.validate(part_model=self.bike)

        boolean = self.bike.property(name="Sale?")
        boolean_filter = PropertyValueFilter(
            property_model=boolean,
            value=False,
            filter_type=FilterType.CONTAINS,
        )

        with self.assertWarns(Warning, msg="Using any filter but EXACT should be warned against"):
            boolean_filter.validate(part_model=self.bike)

        self.assertIn(
            "false",
            boolean_filter.format(),
            msg="Boolean value should have been parsed to a string",
        )

    def test_parse_options(self):
        for options in [
            dict(),
            dict(prefilters=dict()),
            dict(prefilters=dict(property_value="")),
        ]:
            with self.subTest():
                prefilters = PropertyValueFilter.parse_options(options=options)

                self.assertIsInstance(prefilters, list)
                self.assertEqual(0, len(prefilters))

        prefilters = PropertyValueFilter.parse_options(
            dict(prefilters=dict(property_value=self.filter.format()))
        )

        self.assertIsInstance(prefilters, list)
        self.assertEqual(1, len(prefilters))

    def test_write_options(self):
        options_dict = PropertyValueFilter.write_options(filters=[self.filter])

        self.assertIsInstance(options_dict, dict)

        with self.assertRaises(IllegalArgumentError):
            ScopeFilter.write_options(filters=[self.filter])

    def test__eq__(self):
        second_filter = PropertyValueFilter(
            self.bike.property("Gears"), 15, FilterType.GREATER_THAN_EQUAL
        )
        third_filter = PropertyValueFilter(
            self.bike.property("Gears"), 16, FilterType.GREATER_THAN_EQUAL
        )
        fourth_filter = PropertyValueFilter(
            self.bike.property("Total height"), 15, FilterType.GREATER_THAN_EQUAL
        )
        fifth_filter = PropertyValueFilter(
            self.bike.property("Gears"), 15, FilterType.LOWER_THAN_EQUAL
        )

        self.assertEqual(self.filter, second_filter)
        self.assertNotEqual(self.filter, "Coffee filter")
        self.assertNotEqual(self.filter, third_filter)
        self.assertNotEqual(self.filter, fourth_filter)
        self.assertNotEqual(self.filter, fifth_filter)


class TestPropertyFilterAllPropertyTypes(TestBetamax):
    def setUp(self):
        super().setUp()

        self.bike = self.project.product_root_model("Bike")
        self.bike_instance = self.bike.instance()
        self.root = self.project.activity(name=ActivityRootNames.WORKFLOW_ROOT)
        self.wheel = self.project.model(name="Wheel")
        self.test_activity = self.project.create_activity(
            name="__TEST__FILTERS__",
            activity_type=ActivityType.TASK,
        )
        self.wm = self.test_activity.widgets()
        self.prop_test_name = "__PROP TEST"
        self.new_wheel = self.bike_instance.add(model=self.wheel, name="Wheel, Wheel")
        self.new_part = self.bike.add_model(
            name="__TEST_PART__", multiplicity=Multiplicity.ZERO_MANY
        )
        self.filter_types = FilterType.values()

    def tearDown(self):
        self.test_activity.delete()
        self.new_part.delete()
        self.new_wheel.delete()
        super().tearDown()

    def test_text_property_filter_in_grid(self):
        filter_value = "sample"
        test_prop = self.new_part.add_property(
            name=self.prop_test_name, property_type=PropertyType.TEXT_VALUE
        )
        self.bike_instance.add_with_properties(
            model=self.new_part, update_dict={test_prop.id: "sample text"}
        )
        self.bike_instance.add_with_properties(
            model=self.new_part, update_dict={test_prop.id: "just text"}
        )
        with warnings.catch_warnings(record=True) as w:
            for filter_type in self.filter_types:
                prefilter = PropertyValueFilter(
                    property_model=test_prop, value=filter_value, filter_type=filter_type
                )
                widget = self.wm.add_filteredgrid_widget(
                    title=filter_type,
                    part_model=self.new_part,
                    readable_models=list(),
                    writable_models=[test_prop],
                    custom_height=300,
                    parent_instance=self.bike_instance,
                    prefilters=[prefilter],
                )
                self.assertEqual(
                    widget.meta.get("prefilters").get("property_value"),
                    f"{test_prop.id}:{filter_value}:{filter_type}",
                )
            self.assertEqual(len(w), 4)

    def test_multi_test_property_filter_in_grid(self):
        filter_value = "sample"
        test_prop = self.new_part.add_property(
            name=self.prop_test_name, property_type=PropertyType.CHAR_VALUE
        )
        self.bike_instance.add_with_properties(
            model=self.new_part, update_dict={test_prop.id: "sample text"}
        )
        self.bike_instance.add_with_properties(
            model=self.new_part, update_dict={test_prop.id: "just text"}
        )
        with warnings.catch_warnings(record=True) as w:
            for filter_type in self.filter_types:
                prefilter = PropertyValueFilter(
                    property_model=test_prop, value=filter_value, filter_type=filter_type
                )
                widget = self.wm.add_filteredgrid_widget(
                    title=filter_type,
                    part_model=self.new_part,
                    readable_models=list(),
                    writable_models=[test_prop],
                    custom_height=300,
                    parent_instance=self.bike_instance,
                    prefilters=[prefilter],
                )
                self.assertEqual(
                    widget.meta.get("prefilters").get("property_value"),
                    f"{test_prop.id}:{filter_value}:{filter_type}",
                )
            self.assertEqual(len(w), 4)

    def test_int_property_filter_in_grid(self):
        filter_value = 17
        test_prop = self.new_part.add_property(
            name=self.prop_test_name, property_type=PropertyType.INT_VALUE
        )
        self.bike_instance.add_with_properties(model=self.new_part, update_dict={test_prop.id: 15})
        self.bike_instance.add_with_properties(model=self.new_part, update_dict={test_prop.id: 24})
        with warnings.catch_warnings(record=True) as w:
            for filter_type in self.filter_types:
                prefilter = PropertyValueFilter(
                    property_model=test_prop, value=filter_value, filter_type=filter_type
                )
                widget = self.wm.add_filteredgrid_widget(
                    title=filter_type,
                    part_model=self.new_part,
                    readable_models=list(),
                    writable_models=[test_prop],
                    custom_height=300,
                    parent_instance=self.bike_instance,
                    prefilters=[prefilter],
                )
                self.assertEqual(
                    widget.meta.get("prefilters").get("property_value"),
                    f"{test_prop.id}:{filter_value}:{filter_type}",
                )
            self.assertEqual(len(w), 3)

    def test_float_property_filter_in_grid(self):
        filter_value = 22.55
        test_prop = self.new_part.add_property(
            name=self.prop_test_name, property_type=PropertyType.FLOAT_VALUE
        )
        self.bike_instance.add_with_properties(
            model=self.new_part, update_dict={test_prop.id: 15.5}
        )
        self.bike_instance.add_with_properties(
            model=self.new_part, update_dict={test_prop.id: 24.4}
        )
        with warnings.catch_warnings(record=True) as w:
            for filter_type in self.filter_types:
                prefilter = PropertyValueFilter(
                    property_model=test_prop, value=filter_value, filter_type=filter_type
                )
                widget = self.wm.add_filteredgrid_widget(
                    title=filter_type,
                    part_model=self.new_part,
                    readable_models=list(),
                    writable_models=[test_prop],
                    custom_height=300,
                    parent_instance=self.bike_instance,
                    prefilters=[prefilter],
                )
                self.assertEqual(
                    widget.meta.get("prefilters").get("property_value"),
                    f"{test_prop.id}:{filter_value}:{filter_type}",
                )
            self.assertEqual(len(w), 3)

    def test_boolean_property_filter_in_grid(self):
        filter_value = True
        test_prop = self.new_part.add_property(
            name=self.prop_test_name, property_type=PropertyType.BOOLEAN_VALUE
        )
        self.bike_instance.add_with_properties(
            model=self.new_part, update_dict={test_prop.id: False}
        )
        self.bike_instance.add_with_properties(
            model=self.new_part, update_dict={test_prop.id: True}
        )
        with warnings.catch_warnings(record=True) as w:
            for filter_type in self.filter_types:
                prefilter = PropertyValueFilter(
                    property_model=test_prop, value=filter_value, filter_type=filter_type
                )
                widget = self.wm.add_filteredgrid_widget(
                    title=filter_type,
                    part_model=self.new_part,
                    readable_models=list(),
                    writable_models=[test_prop],
                    custom_height=300,
                    parent_instance=self.bike_instance,
                    prefilters=[prefilter],
                )
                self.assertEqual(
                    widget.meta.get("prefilters").get("property_value"),
                    "{}:{}:{}".format(test_prop.id, str(filter_value).lower(), filter_type),
                )
            self.assertEqual(len(w), 4)

    def test_date_property_filter_in_grid(self):
        filter_value = str(datetime.date(2021, 5, 4))
        test_prop = self.new_part.add_property(
            name=self.prop_test_name, property_type=PropertyType.DATE_VALUE
        )
        self.bike_instance.add_with_properties(
            model=self.new_part, update_dict={test_prop.id: str(datetime.date(2021, 5, 2))}
        )
        self.bike_instance.add_with_properties(
            model=self.new_part, update_dict={test_prop.id: str(datetime.date(2021, 5, 8))}
        )
        with warnings.catch_warnings(record=True) as w:
            for filter_type in self.filter_types:
                prefilter = PropertyValueFilter(
                    property_model=test_prop, value=filter_value, filter_type=filter_type
                )
                widget = self.wm.add_filteredgrid_widget(
                    title=filter_type,
                    part_model=self.new_part,
                    readable_models=list(),
                    writable_models=[test_prop],
                    custom_height=300,
                    parent_instance=self.bike_instance,
                    prefilters=[prefilter],
                )
                self.assertEqual(
                    widget.meta.get("prefilters").get("property_value"),
                    "{}:{}:{}".format(test_prop.id, filter_value, filter_type),
                )
            self.assertEqual(len(w), 3)

    def test_link_property_filter_in_grid(self):
        filter_value = "nl"
        test_prop = self.new_part.add_property(
            name=self.prop_test_name, property_type=PropertyType.LINK_VALUE
        )
        self.bike_instance.add_with_properties(
            model=self.new_part, update_dict={test_prop.id: "https://ke-chain.com"}
        )
        self.bike_instance.add_with_properties(
            model=self.new_part, update_dict={test_prop.id: "https://ke-chain.nl"}
        )
        with warnings.catch_warnings(record=True) as w:
            for filter_type in self.filter_types:
                prefilter = PropertyValueFilter(
                    property_model=test_prop, value=filter_value, filter_type=filter_type
                )
                widget = self.wm.add_filteredgrid_widget(
                    title=filter_type,
                    part_model=self.new_part,
                    readable_models=list(),
                    writable_models=[test_prop],
                    custom_height=300,
                    parent_instance=self.bike_instance,
                    prefilters=[prefilter],
                )
                self.assertEqual(
                    widget.meta.get("prefilters").get("property_value"),
                    "{}:{}:{}".format(test_prop.id, filter_value, filter_type),
                )
            self.assertEqual(len(w), 4)

    def test_single_select_property_filter_in_grid(self):
        filter_value = "apples"
        test_prop = self.new_part.add_property(
            name=self.prop_test_name,
            options=dict(value_choices=["apples", "oranges", "bananas", "lemons"]),
            property_type=PropertyType.SINGLE_SELECT_VALUE,
        )
        self.bike_instance.add_with_properties(
            model=self.new_part, update_dict={test_prop.id: "apples"}
        )
        self.bike_instance.add_with_properties(
            model=self.new_part, update_dict={test_prop.id: "bananas"}
        )
        with warnings.catch_warnings(record=True) as w:
            for filter_type in self.filter_types:
                prefilter = PropertyValueFilter(
                    property_model=test_prop, value=filter_value, filter_type=filter_type
                )
                widget = self.wm.add_filteredgrid_widget(
                    title=filter_type,
                    part_model=self.new_part,
                    readable_models=list(),
                    writable_models=[test_prop],
                    custom_height=300,
                    parent_instance=self.bike_instance,
                    prefilters=[prefilter],
                )
                self.assertEqual(
                    widget.meta.get("prefilters").get("property_value"),
                    "{}:{}:{}".format(test_prop.id, filter_value, filter_type),
                )
            self.assertEqual(len(w), 4)

    def test_multi_select_property_filter_in_grid(self):
        filter_value = "apples"
        test_prop = self.new_part.add_property(
            name=self.prop_test_name,
            options=dict(value_choices=["apples", "oranges", "bananas", "lemons"]),
            property_type=PropertyType.MULTI_SELECT_VALUE,
        )
        self.bike_instance.add_with_properties(
            model=self.new_part, update_dict={test_prop.id: ["apples", "bananas"]}
        )
        self.bike_instance.add_with_properties(
            model=self.new_part, update_dict={test_prop.id: ["bananas", "lemons"]}
        )
        with warnings.catch_warnings(record=True) as w:
            for filter_type in self.filter_types:
                prefilter = PropertyValueFilter(
                    property_model=test_prop, value=filter_value, filter_type=filter_type
                )
                widget = self.wm.add_filteredgrid_widget(
                    title=filter_type,
                    part_model=self.new_part,
                    readable_models=list(),
                    writable_models=[test_prop],
                    custom_height=300,
                    parent_instance=self.bike_instance,
                    prefilters=[prefilter],
                )
                self.assertEqual(
                    widget.meta.get("prefilters").get("property_value"),
                    "{}:{}:{}".format(test_prop.id, filter_value, filter_type),
                )
            self.assertEqual(len(w), 4)

    def test_part_reference_property_filter_in_grid(self):
        filter_value = "Rear Wheel"
        test_prop = self.new_part.add_property(
            name=self.prop_test_name,
            default_value=self.wheel,
            property_type=PropertyType.REFERENCES_VALUE,
        )
        self.bike_instance.add_with_properties(
            model=self.new_part,
            update_dict={test_prop.id: [self.project.part(name="Front Wheel")]},
        )
        self.bike_instance.add_with_properties(
            model=self.new_part,
            update_dict={
                test_prop.id: [
                    self.project.part(name="Front Wheel"),
                    self.project.part(name="Rear Wheel"),
                ]
            },
        )
        with warnings.catch_warnings(record=True) as w:
            for filter_type in self.filter_types:
                prefilter = PropertyValueFilter(
                    property_model=test_prop, value=filter_value, filter_type=filter_type
                )
                widget = self.wm.add_filteredgrid_widget(
                    title=filter_type,
                    part_model=self.new_part,
                    readable_models=list(),
                    writable_models=[test_prop],
                    custom_height=300,
                    parent_instance=self.bike_instance,
                    prefilters=[prefilter],
                )
                self.assertEqual(
                    widget.meta.get("prefilters").get("property_value"),
                    "{}:{}:{}".format(test_prop.id, urllib.parse.quote(filter_value), filter_type),
                )
            self.assertEqual(len(w), 4)

    def test_part_reference_property_filter_in_grid_with_commas(self):
        filter_value = self.new_wheel.name
        test_prop = self.new_part.add_property(
            name=self.prop_test_name,
            default_value=self.wheel,
            property_type=PropertyType.REFERENCES_VALUE,
        )
        self.bike_instance.add_with_properties(
            model=self.new_part,
            update_dict={test_prop.id: [self.project.part(name="Front Wheel")]},
        )
        self.bike_instance.add_with_properties(
            model=self.new_part,
            update_dict={
                test_prop.id: [
                    self.project.part(name="Front Wheel"),
                    self.project.part(name="Rear Wheel"),
                ]
            },
        )
        with warnings.catch_warnings(record=True) as w:
            for filter_type in self.filter_types:
                prefilter = PropertyValueFilter(
                    property_model=test_prop, value=filter_value, filter_type=filter_type
                )
                widget = self.wm.add_filteredgrid_widget(
                    title=filter_type,
                    part_model=self.new_part,
                    readable_models=list(),
                    writable_models=[test_prop],
                    custom_height=300,
                    parent_instance=self.bike_instance,
                    prefilters=[prefilter],
                )
                self.assertEqual(
                    widget.meta.get("prefilters").get("property_value"),
                    "{}:{}:{}".format(test_prop.id, urllib.parse.quote(filter_value), filter_type),
                )
            self.assertEqual(len(w), 4)

    def test_activity_reference_property_filter_in_grid(self):
        filter_value = "Specify wheel diameter"
        test_prop = self.new_part.add_property(
            name=self.prop_test_name, property_type=PropertyType.ACTIVITY_REFERENCES_VALUE
        )
        self.bike_instance.add_with_properties(
            model=self.new_part,
            update_dict={test_prop.id: [self.project.activity(name="Subprocess")]},
        )
        self.bike_instance.add_with_properties(
            model=self.new_part,
            update_dict={
                test_prop.id: [
                    self.project.activity(name="Specify wheel diameter"),
                    self.project.activity(name="Subprocess"),
                ]
            },
        )
        with warnings.catch_warnings(record=True) as w:
            for filter_type in self.filter_types:
                prefilter = PropertyValueFilter(
                    property_model=test_prop, value=filter_value, filter_type=filter_type
                )
                widget = self.wm.add_filteredgrid_widget(
                    title=filter_type,
                    part_model=self.new_part,
                    readable_models=list(),
                    writable_models=[test_prop],
                    custom_height=300,
                    parent_instance=self.bike_instance,
                    prefilters=[prefilter],
                )
                self.assertEqual(
                    widget.meta.get("prefilters").get("property_value"),
                    "{}:{}:{}".format(test_prop.id, urllib.parse.quote(filter_value), filter_type),
                )
            self.assertEqual(len(w), 4)

    def test_user_reference_property_filter_in_grid(self):
        filter_value = "Test"
        test_prop = self.new_part.add_property(
            name=self.prop_test_name, property_type=PropertyType.USER_REFERENCES_VALUE
        )
        self.bike_instance.add_with_properties(
            model=self.new_part,
            update_dict={test_prop.id: [self.client.user(username="superuser")]},
        )
        self.bike_instance.add_with_properties(
            model=self.new_part,
            update_dict={
                test_prop.id: [
                    self.client.user(username="testuser"),
                    self.client.user(username="superuser"),
                ]
            },
        )
        with warnings.catch_warnings(record=True) as w:
            for filter_type in self.filter_types:
                prefilter = PropertyValueFilter(
                    property_model=test_prop, value=filter_value, filter_type=filter_type
                )
                widget = self.wm.add_filteredgrid_widget(
                    title=filter_type,
                    part_model=self.new_part,
                    readable_models=list(),
                    writable_models=[test_prop],
                    custom_height=300,
                    parent_instance=self.bike_instance,
                    prefilters=[prefilter],
                )
                self.assertEqual(
                    widget.meta.get("prefilters").get("property_value"),
                    "{}:{}:{}".format(test_prop.id, filter_value, filter_type),
                )
            self.assertEqual(len(w), 4)

    def test_scope_reference_property_filter_in_grid(self):
        filter_value = "Bike"
        test_prop = self.new_part.add_property(
            name=self.prop_test_name, property_type=PropertyType.SCOPE_REFERENCES_VALUE
        )
        self.bike_instance.add_with_properties(
            model=self.new_part,
            update_dict={test_prop.id: [self.client.scope("Cannondale Project")]},
        )
        self.bike_instance.add_with_properties(
            model=self.new_part,
            update_dict={
                test_prop.id: [
                    self.client.scope("Cannondale Project"),
                    self.client.scope("Bike Project"),
                ]
            },
        )
        with warnings.catch_warnings(record=True) as w:
            for filter_type in self.filter_types:
                prefilter = PropertyValueFilter(
                    property_model=test_prop, value=filter_value, filter_type=filter_type
                )
                widget = self.wm.add_filteredgrid_widget(
                    title=filter_type,
                    part_model=self.new_part,
                    readable_models=list(),
                    writable_models=[test_prop],
                    custom_height=300,
                    parent_instance=self.bike_instance,
                    prefilters=[prefilter],
                )
                self.assertEqual(
                    widget.meta.get("prefilters").get("property_value"),
                    "{}:{}:{}".format(test_prop.id, filter_value, filter_type),
                )
            self.assertEqual(len(w), 4)

    def test_part_reference_property_prefilter(self):
        filter_type = FilterType.GREATER_THAN_EQUAL
        filter_value = 4.2
        filter_property_model = self.wheel.property("Tire Thickness")
        test_prop = self.new_part.add_property(
            name=self.prop_test_name,
            default_value=self.wheel,
            property_type=PropertyType.REFERENCES_VALUE,
        )
        prefilter = PropertyValueFilter(
            property_model=filter_property_model, value=filter_value, filter_type=filter_type
        )
        test_prop.set_prefilters(prefilters=[prefilter])
        self.assertEqual(
            test_prop._options.get("prefilters").get("property_value"),
            f"{filter_property_model.id}:{filter_value}:{filter_type}",
        )

    def test_activity_reference_property_prefilter(self):
        test_prop = self.new_part.add_property(
            name=self.prop_test_name, property_type=PropertyType.SCOPE_REFERENCES_VALUE
        )
        prefilters = [
            ScopeFilter(tag="bike"),
            ScopeFilter(tag="bmx"),
            ScopeFilter(progress_gte=1.0),
        ]
        test_prop.set_prefilters(prefilters=prefilters)
        self.assertEqual(test_prop._options.get("prefilters").get("tags__contains"), "bike,bmx")
        self.assertEqual(test_prop._options.get("prefilters").get("progress__gte"), 1.0)


class BaseTest:
    class _TestScopeFilter(TestCase):

        VALUE = None
        VALUE_2 = None
        INVALID_VALUE = None
        FIELD = None
        ATTR = None

        @classmethod
        def setUpClass(cls) -> None:
            super().setUpClass()
            if cls.VALUE == cls.VALUE_2:  # pragma: no cover
                raise ValueError(f"Test needs 2 different values ({cls.VALUE}, {cls.VALUE_2})")

        def setUp(self) -> None:
            super().setUp()
            self.filter = ScopeFilter(**{self.ATTR: self.VALUE})

        def test__repr__(self):
            representation = self.filter.__repr__()

            self.assertIsInstance(representation, str)

        def test__eq__(self):
            second_filter = ScopeFilter(**{self.ATTR: self.VALUE})
            third_filter = ScopeFilter(**{self.ATTR: self.VALUE_2})

            self.assertEqual(self.filter, second_filter)
            self.assertNotEqual(self.filter, third_filter)

        def test_write_options(self):
            filter_2 = ScopeFilter(**{self.ATTR: self.VALUE_2})
            filters = [self.filter, filter_2]

            options_dict = ScopeFilter.write_options(filters=filters)

            self.assertIsInstance(options_dict, dict)

            with self.assertRaises(IllegalArgumentError):
                PropertyValueFilter.write_options(filters=[self.filter])

        def test_parse_options(self):
            options = ScopeFilter.write_options(filters=[self.filter])

            scope_filters = ScopeFilter.parse_options(options=options)

            self.assertTrue(scope_filters)
            self.assertIsInstance(scope_filters, list)
            self.assertTrue(all(isinstance(sf, ScopeFilter) for sf in scope_filters))

        def test_creation(self):
            if self.INVALID_VALUE is not None:
                with self.assertRaises(IllegalArgumentError):
                    ScopeFilter(**{self.ATTR: self.INVALID_VALUE})

            with self.assertRaises(IllegalArgumentError):
                ScopeFilter(tag="My scope tag", status=ScopeStatus.ACTIVE)


class TestScopeFilterName(BaseTest._TestScopeFilter):
    VALUE = "My project"
    VALUE_2 = "Not my project"
    INVALID_VALUE = 3
    FIELD = "name__icontains"
    ATTR = "name"


class TestScopeFilterStatus(BaseTest._TestScopeFilter):
    VALUE = ScopeStatus.CLOSED
    VALUE_2 = ScopeStatus.ACTIVE
    INVALID_VALUE = "Just a fleshwound"
    FIELD = "status__in"
    ATTR = "status"


class TestScopeFilterDueDateGTE(BaseTest._TestScopeFilter):
    VALUE = TIMESTAMP
    VALUE_2 = TIMESTAMP_2
    INVALID_VALUE = 3
    FIELD = "due_date__gte"
    ATTR = "due_date_gte"


class TestScopeFilterDueDateLTE(BaseTest._TestScopeFilter):
    VALUE = TIMESTAMP
    VALUE_2 = TIMESTAMP_2
    INVALID_VALUE = 3
    FIELD = "due_date__lte"
    ATTR = "due_date_lte"


class TestScopeFilterStartDateGTE(BaseTest._TestScopeFilter):
    VALUE = TIMESTAMP
    VALUE_2 = TIMESTAMP_2
    INVALID_VALUE = 3
    FIELD = "start_date__gte"
    ATTR = "start_date_gte"


class TestScopeFilterStartDateLTE(BaseTest._TestScopeFilter):
    VALUE = TIMESTAMP
    VALUE_2 = TIMESTAMP_2
    INVALID_VALUE = 3
    FIELD = "start_date__lte"
    ATTR = "start_date_lte"


class TestScopeFilterProgressGTE(BaseTest._TestScopeFilter):
    VALUE = 0.2
    VALUE_2 = 0.3
    INVALID_VALUE = "completed"
    FIELD = "progress__gte"
    ATTR = "progress_gte"


class TestScopeFilterProgressLTE(BaseTest._TestScopeFilter):
    VALUE = 0.2
    VALUE_2 = 0.3
    INVALID_VALUE = "completed"
    FIELD = "progress__lte"
    ATTR = "progress_lte"


class TestScopeFilterTag(BaseTest._TestScopeFilter):
    VALUE = "Calculation"
    VALUE_2 = "Not a calculation"
    INVALID_VALUE = 3
    FIELD = "tags__contains"
    ATTR = "tag"


class TestScopeFilterTeam(BaseTest._TestScopeFilter):
    VALUE = "a2a0631b-d771-4807-bd22-c5ccde581e79"
    VALUE_2 = "9bab05df-5597-4a7a-af65-73cc65fee22a"
    INVALID_VALUE = 3
    FIELD = "team__in"
    ATTR = "team"


class TestScopeFilterUnknown(BaseTest._TestScopeFilter):
    VALUE = 1
    VALUE_2 = 2
    INVALID_VALUE = None
    FIELD = "new_filter__type"
    ATTR = "new_filter_type"



from unittest import TestCase

from pykechain.exceptions import IllegalArgumentError
from pykechain.models import Activity
from pykechain.models.tags import TagsMixin
from tests.classes import TestBetamax


class TestTags(TestCase):
    tags = ["one", "two", "three"]

    def setUp(self):
        self.obj = ConcreteTagsBase()
        self.obj._tags = self.tags

    def tearDown(self):
        pass

    def test_edit(self):
        self.assertIsNone(self.obj.edit(tags=self.tags))

    def test_getter(self):
        self.assertTrue(len(self.tags) == len(self.obj.tags))
        self.assertTrue(sorted(self.tags) == sorted(self.obj.tags))

    def test_setter(self):
        new_tags = ["1", "2", "3"]
        self.obj.tags = new_tags

        self.assertTrue(len(new_tags) == len(self.obj.tags))
        self.assertTrue(sorted(new_tags) == sorted(self.obj.tags))

    def test_setter_duplicate(self):
        self.obj.add_tag(tag="one")
        self.assertTrue(len(self.obj.tags) == 3)

    def test_setter_none(self):
        self.obj.tags = None

        self.assertIsInstance(self.obj.tags, list)
        self.assertTrue(len(self.obj.tags) == 0)

    def test_setter_not_an_iterable(self):
        with self.assertRaises(IllegalArgumentError):
            self.obj.tags = "four"

    def test_setter_not_a_string(self):
        with self.assertRaises(ValueError):
            self.obj.tags = [1, 2, 3]

    def test_remove_tag(self):
        self.obj.remove_tag(tag="two")

        self.assertNotIn("two", self.obj.tags)

    def test_remove_tag_failure(self):
        with self.assertRaises(IllegalArgumentError):
            self.obj.remove_tag(tag="just a flesh wound")

    def test_add_tag(self):
        self.obj.add_tag(tag="four")

        self.assertIn("four", self.obj.tags)

    def test_has_tag_true(self):
        self.assertTrue(self.obj.has_tag(tag="three"))

    def test_has_tag_false(self):
        self.assertFalse(self.obj.has_tag(tag="ten"))


class TestTagsScope(TestBetamax):
    def test_scope_tags(self):
        """test to retrieve the tags for a scope"""
        # setup
        saved_tags = self.project.tags

        # test
        scope_tags = ["a", "list", "of-tags"]
        self.project.edit(tags=scope_tags)
        self.assertListEqual(scope_tags, self.project.tags)

        self.assertTrue(len(scope_tags) == len(self.project.tags))
        self.assertTrue(sorted(scope_tags) == sorted(self.project.tags))

        # teardown
        self.project.edit(tags=saved_tags)

    def test_scope_tags_may_be_emptied(self):
        # setup
        saved_tags = self.project.tags

        # test
        self.project.edit(tags=[])
        self.assertListEqual(self.project.tags, [])

        # teardown
        self.project.edit(tags=saved_tags)


class TestTagsActivity(TestBetamax):
    tags = ["one", "two", "three"]

    def setUp(self):
        super().setUp()
        self.task = self.project.activity(name="SubTask")  # type: Activity

    def tearDown(self):
        self.task.tags = None
        super().tearDown()

    def test_activity_tags(self):
        # setup
        new_tags = ["four", "five", "six"]
        self.task.tags = new_tags
        reloaded = self.project.activity(pk=self.task.id)

        self.assertEqual(len(new_tags), len(reloaded.tags))
        self.assertListEqual(sorted(new_tags), sorted(reloaded.tags))

    def test_activity_tags_may_be_emptied(self):
        # setup
        self.task.tags = []
        reloaded_task = self.project.activity(pk=self.task.id)

        self.assertEqual(0, len(reloaded_task.tags))


class ConcreteTagsBase(TagsMixin):
    """
    Dummy implementation of the edit method allows for instantiation of the TagsMixin class.
    """

    def edit(self, tags=None, *args, **kwargs):
        return None



#!/usr/bin/env python

"""
Debugging script returning the environment etc
"""

import os
import platform
import sys
import time
from pprint import pprint


def main(*args, **kwargs):
    start = time.time()
    foo, bar = "cow says", "moo!"
    print(">>> starting on {}".format(time.strftime("%c", time.gmtime(start))))
    print(f"--- arguments: {args}")
    print(f"--- keyword arguments: {kwargs}")
    print(f"--- local variables dict: {dir()}")
    pprint(f"--- environment: {os.environ}", width=120)
    pprint(f"--- platform details (uname): {platform.uname()}", width=120)
    pprint(f"--- python paths: {sys.path}", width=120)
    mark_1 = time.time()
    print(
        "<<< intermediate timing on {}, duration: {}".format(
            time.strftime("%c", time.gmtime(mark_1)), mark_1 - start
        )
    )
    print("--- python pip packages installed \n{}".format(os.popen("pip freeze").readlines()))
    try:
        import pykechain

        print(f"--- importing pykechain version: {pykechain.version}")
    except ImportError:
        print(r"/!\ pykechain is not installed, please install pykechain version 1.12 or later")

    mark_2 = time.time()

    from envparse import env

    if (
        not env("KECHAIN_URL", None)
        or not env("KECHAIN_TOKEN", None)
        or not env("KECHAIN_SCOPE_ID", None)
    ):
        print(
            r"/!\ cannot interact using pykechain as environment variables KECHAIN_URL,"
            r" KECHAIN_TOKEN and "
            "KECHAIN_SCOPE_ID are not set"
        )
    else:
        project = pykechain.get_project()
        mark_3 = time.time()
        print(
            "--- retrieving project '{}' from kechain (url: {})".format(
                project.name, env("KECHAIN_URL")
            )
        )
        print(f"--- {len(project.activities())} activities, {len(project.parts())} parts")
        print("--- retrieving as user: '{}".format(env("KECHAIN_USERNAME")))
        print(f"<<< interaction with ke-chain took: {mark_3-mark_2}s")

    end = time.time()
    print(
        "<<< ending on {}, duration: {}".format(time.strftime("%c", time.gmtime(end)), end - start)
    )


if __name__ == "__main__":
    sys.exit(main())



import os
from datetime import datetime, date, timezone

from pykechain.enums import FilterType, Multiplicity, PropertyType, ScopeStatus
from pykechain.exceptions import NotFoundError, IllegalArgumentError
from pykechain.models.value_filter import ScopeFilter
from tests.classes import TestBetamax


class TestPartsCopyMove(TestBetamax):
    project_root = os.path.dirname(os.path.dirname(os.path.abspath(__file__)).replace("\\", "/"))

    def setUp(self):
        super().setUp()
        self.base = self.project.model(name__startswith="Catalog")
        self.cross_scope_project = self.client.scope(ref="cannondale-project")
        self.cross_scope_bike = self.cross_scope_project.model(ref="cannondale-bike")
        self.wheel = self.project.model(name="Wheel")
        self.front_wheel, self.rear_wheel = self.project.part(
            name="Front Wheel"
        ), self.project.part(name="Rear Wheel")
        self.specify_wheel_diameter = self.project.activity(name="Specify wheel diameter")
        self.model_to_be_copied = self.project.create_model(
            parent=self.base,
            name=f"__Model to be Copied @ {str(datetime.now())} [TEST]",
            multiplicity=Multiplicity.ONE_MANY,
        )  # type Part
        self.p_char_name = "__Property single text"
        self.p_multi_name = "__Property multi text"
        self.p_bool_name = "__Property boolean"
        self.p_int_name = "__Property integer"
        self.p_float_name = "__Property float"
        self.p_date_name = "__Property date"
        self.p_datetime_name = "__Property datetime"
        self.p_attach_name = "__Property attachment"
        self.p_link_name = "__Property link"
        self.p_single_select_name = "__Property single select"
        self.p_multi_select_name = "__Property multi select"
        self.p_part_reference_name = "__Property part reference"
        self.p_part_reference_name_cross_scope = "__Property part reference cross scope"
        self.p_geo_json_name = "__Property geographical"
        self.p_weather_name = "__Property weather"
        self.p_activity_reference_name = "__Property activity reference"
        self.p_user_reference_name = "__Property user reference"
        self.p_project_reference_name = "__Property project reference"
        self.p_char_value = "test value"
        self.p_multi_value = "test multi text value"
        self.p_bool_value = True
        self.p_int_value = 33
        self.p_float_value = 42.42
        self.p_date_value = str(date(2021, 12, 15))
        self.p_datetime_value = str(datetime(1991, 4, 12, 13, 5, 5).isoformat())
        self.p_attach_value = self.project_root + "/requirements.txt"
        self.p_link_value = "http://ke-chain.com"
        self.p_select_value_choices = ["apples", "oranges", "bananas", "lemons"]
        self.p_single_select_value = "bananas"
        self.p_multi_select_value = ["apples", "bananas"]
        self.test_manager = self.client.user(username="testmanager")
        self.test_supervisor = self.client.user(username="testsupervisor")
        self.p_user_reference_value = [self.test_manager.id, self.test_supervisor.id]
        self.p_geojson_value = {
            "type": "FeatureCollection",
            "features": [
                {
                    "type": "Feature",
                    "geometry": {"type": "Point", "coordinates": [4.390912, 51.9929856]},
                    "properties": {
                        "zoom": 16,
                        "address": "Rijksstraatweg 47, 2645 BS Delfgauw, Netherlands",
                    },
                }
            ],
        }
        self.p_geojson_value_options = {
            "representations": [
                {"rtype": "geoCoordinate", "config": {"geoCoordinate": "approx_address"}}
            ]
        }
        self.p_weather_value = {
            "current": {
                "uv": 1,
                "cloud": 0,
                "is_day": 1,
                "temp_c": 12,
                "temp_f": 53.6,
                "vis_km": 10,
                "gust_kph": 11.2,
                "gust_mph": 6.9,
                "humidity": 88,
                "wind_dir": "NE",
                "wind_kph": 0,
                "wind_mph": 0,
                "condition": {
                    "code": 1000,
                    "icon": "//cdn.weatherapi.com/weather/64x64/day/113.png",
                    "text": "Sunny",
                },
                "precip_in": 0,
                "precip_mm": 0,
                "vis_miles": 6,
                "feelslike_c": 11.9,
                "feelslike_f": 53.5,
                "pressure_in": 30.42,
                "pressure_mb": 1030,
                "wind_degree": 52,
                "last_updated": "2021-09-21 09:30",
                "last_updated_epoch": 1632209400,
            },
            "location": {
                "lat": 51.99,
                "lon": 4.39,
                "name": "Delfgauw",
                "tz_id": "Europe/Amsterdam",
                "region": "South Holland",
                "country": "Netherlands",
                "localtime": "2021-09-21 9:39",
                "localtime_epoch": 1632209956,
            },
        }
        self.sub_part1 = self.model_to_be_copied.add_model(
            name="__subpart 1", multiplicity=Multiplicity.ONE
        )
        self.sub_part2 = self.model_to_be_copied.add_model(
            name="__subpart 2", multiplicity=Multiplicity.ZERO_MANY
        )
        self.model_to_be_copied.add_property(
            name=self.p_char_name,
            description="Description of Property single text",
            unit="mm",
            default_value=self.p_char_value,
            property_type=PropertyType.CHAR_VALUE,
        )
        self.model_to_be_copied.add_property(
            name=self.p_multi_name,
            unit="mm",
            default_value=self.p_multi_value,
            property_type=PropertyType.TEXT_VALUE,
        )
        self.model_to_be_copied.add_property(
            name=self.p_bool_name,
            default_value=self.p_bool_value,
            property_type=PropertyType.BOOLEAN_VALUE,
        )
        self.model_to_be_copied.add_property(
            name=self.p_int_name,
            default_value=self.p_int_value,
            property_type=PropertyType.INT_VALUE,
        )
        self.model_to_be_copied.add_property(
            name=self.p_float_name,
            default_value=self.p_float_value,
            property_type=PropertyType.FLOAT_VALUE,
        )
        self.model_to_be_copied.add_property(
            name=self.p_date_name,
            default_value=self.p_date_value,
            property_type=PropertyType.DATE_VALUE,
        )
        self.model_to_be_copied.add_property(
            name=self.p_datetime_name,
            default_value=self.p_datetime_value,
            property_type=PropertyType.DATETIME_VALUE,
        )
        self.attachment_property = self.model_to_be_copied.add_property(
            name=self.p_attach_name, property_type=PropertyType.ATTACHMENT_VALUE
        )
        self.attachment_property.upload(self.p_attach_value)

        self.model_to_be_copied.add_property(
            name=self.p_link_name,
            default_value=self.p_link_value,
            property_type=PropertyType.LINK_VALUE,
        )
        self.model_to_be_copied.add_property(
            name=self.p_single_select_name,
            options=dict(value_choices=self.p_select_value_choices),
            default_value=self.p_single_select_value,
            property_type=PropertyType.SINGLE_SELECT_VALUE,
        )
        self.model_to_be_copied.add_property(
            name=self.p_multi_select_name,
            default_value=self.p_multi_select_value,
            options=dict(value_choices=self.p_select_value_choices),
            property_type=PropertyType.MULTI_SELECT_VALUE,
        )
        self.model_to_be_copied.add_property(
            name=self.p_part_reference_name,
            default_value=self.sub_part2,
            property_type=PropertyType.REFERENCES_VALUE,
        )
        self.model_to_be_copied.add_property(
            name=self.p_part_reference_name_cross_scope,
            default_value=self.wheel,
            property_type=PropertyType.REFERENCES_VALUE,
        )
        self.model_to_be_copied.add_property(
            name=self.p_geo_json_name,
            property_type=PropertyType.GEOJSON_VALUE,
            default_value=self.p_geojson_value,
            value_options=self.p_geojson_value_options,
        )
        self.model_to_be_copied.add_property(
            name=self.p_weather_name,
            property_type=PropertyType.WEATHER_VALUE,
            default_value=self.p_weather_value,
        )
        self.model_to_be_copied.add_property(
            name=self.p_activity_reference_name,
            property_type=PropertyType.ACTIVITY_REFERENCES_VALUE,
            default_value=[self.specify_wheel_diameter.id],
        )
        self.model_to_be_copied.add_property(
            name=self.p_user_reference_name,
            property_type=PropertyType.USER_REFERENCES_VALUE,
            default_value=self.p_user_reference_value,
        )
        self.model_to_be_copied.add_property(
            name=self.p_project_reference_name,
            property_type=PropertyType.SCOPE_REFERENCES_VALUE,
            default_value=[self.project.id],
        )
        self.sub_part2.add_property(
            name=self.p_bool_name,
            default_value=False,
            property_type=PropertyType.BOOLEAN_VALUE,
        )
        self.sub_part2.add_property(
            name=self.p_int_name,
            default_value=3,
            property_type=PropertyType.INT_VALUE,
        )

        self.model_to_be_copied.property(self.p_part_reference_name).set_excluded_propmodels(
            [self.sub_part2.property(self.p_bool_name)]
        )
        self.model_to_be_copied.property(self.p_part_reference_name).set_prefilters(
            property_models=[self.sub_part2.property(self.p_int_name)],
            values=[3],
            filters_type=[FilterType.GREATER_THAN_EQUAL],
        )
        self.model_to_be_copied.property(self.p_project_reference_name).set_prefilters(
            [ScopeFilter(tag='bike'), ScopeFilter(status=ScopeStatus.ACTIVE)]
        )

        self.model_target_parent = self.project.model(ref="bike")
        self.instance_to_be_copied = self.model_to_be_copied.instances()[0]
        self.sub_part2_instance = self.instance_to_be_copied.add(model=self.sub_part2)
        self.instance_to_be_copied.property(name=self.p_part_reference_name).value = [
            self.sub_part2_instance
        ]
        self.instance_to_be_copied.property(name=self.p_part_reference_name_cross_scope).value = [
            self.front_wheel
        ]
        self.instance_target_parent = self.model_to_be_copied.parent().instances()[0]
        self.dump_part = None

    def tearDown(self):
        self.model_to_be_copied.delete()

        if self.dump_part:
            self.dump_part.delete()
        super().tearDown()

    def test_copy_part_model_given_name_include_children(self):
        model_target_parent = self.project.model("Bike")
        copied_model = self.model_to_be_copied.copy(
            target_parent=model_target_parent,
            name="__Copied model under Bike",
            include_children=True,
            include_instances=False,
        )
        copied_model.refresh()
        copied_model.populate_descendants()
        self.dump_part = copied_model

        # testing
        self.assertTrue(copied_model)
        self.assertEqual(len(copied_model.properties), len(self.model_to_be_copied.properties))
        self.assertEqual(
            copied_model.property(name=self.p_char_name)._json_data["description"],
            "Description of Property single text",
        )
        self.assertEqual(copied_model.property(name=self.p_char_name)._json_data["unit"], "mm")
        self.assertEqual(copied_model.property(name=self.p_int_name).value, self.p_int_value)
        self.assertEqual(copied_model.property(name=self.p_char_name).value, self.p_char_value)
        self.assertEqual(copied_model.property(name=self.p_multi_name).value, self.p_multi_value)
        self.assertEqual(copied_model.property(name=self.p_int_name).value, self.p_int_value)
        self.assertEqual(copied_model.property(name=self.p_float_name).value, self.p_float_value)
        self.assertEqual(copied_model.property(name=self.p_bool_name).value, self.p_bool_value)
        self.assertEqual(copied_model.property(name=self.p_date_name).value, self.p_date_value)
        self.assertIn(
            self.p_datetime_value, copied_model.property(name=self.p_datetime_name).value
        )
        self.assertIn(copied_model.property(name=self.p_attach_name).filename, self.p_attach_value)
        self.assertEqual(copied_model.property(name=self.p_link_name).value, self.p_link_value)
        self.assertEqual(
            copied_model.property(name=self.p_single_select_name).value, self.p_single_select_value
        )
        self.assertEqual(
            copied_model.property(name=self.p_multi_select_name).value, self.p_multi_select_value
        )
        list_of_user_ids = [
            user.id for user in copied_model.property(name=self.p_user_reference_name).value
        ]
        self.assertIn(
            self.sub_part2.name,
            [part.name for part in copied_model.property(name=self.p_part_reference_name).value],
        )
        self.assertNotEqual(
            self.sub_part2.id, copied_model.property(name=self.p_part_reference_name).value[0].id
        )
        self.assertIn(
            self.wheel.id,
            [
                part.id
                for part in copied_model.property(
                    name=self.p_part_reference_name_cross_scope
                ).value
            ],
        )
        self.assertDictEqual(
            self.p_geojson_value, copied_model.property(name=self.p_geo_json_name).value
        )
        self.assertDictEqual(
            self.p_weather_value, copied_model.property(name=self.p_weather_name).value
        )
        self.assertIn(
            self.specify_wheel_diameter.id,
            [act.id for act in copied_model.property(name=self.p_activity_reference_name).value],
        )
        self.assertIn(
            self.project.id,
            [
                scope.id
                for scope in copied_model.property(name=self.p_project_reference_name).value
            ],
        )
        self.assertIn(self.test_manager.id, list_of_user_ids)
        self.assertIn(self.test_supervisor.id, list_of_user_ids)
        self.assertEqual(
            copied_model.property(name=self.p_single_select_name).options,
            self.p_select_value_choices,
        )
        self.assertEqual(
            copied_model.property(name=self.p_multi_select_name).options,
            self.p_select_value_choices,
        )
        self.assertEqual(len(copied_model._cached_children), 2)

    def test_copy_part_model_include_instances(self):
        model_target_parent = self.project.model("Bike")
        self.dump_part = self.model_to_be_copied.copy(
            target_parent=model_target_parent,
            name="__Copied model under Bike",
            include_children=True,
            include_instances=True,
        )
        bike_part = self.project.part("Bike")

        copied_instance = [child for child in bike_part.children() if "Copied" in child.name][0]
        copied_instance.populate_descendants()

        # testing
        self.assertTrue(copied_instance)
        self.assertEqual(len(copied_instance.properties), len(self.model_to_be_copied.properties))
        self.assertEqual(copied_instance.property(name=self.p_char_name).value, self.p_char_value)
        self.assertEqual(
            copied_instance.property(name=self.p_multi_name).value, self.p_multi_value
        )
        self.assertEqual(copied_instance.property(name=self.p_int_name).value, self.p_int_value)
        self.assertEqual(
            copied_instance.property(name=self.p_float_name).value, self.p_float_value
        )
        self.assertEqual(copied_instance.property(name=self.p_bool_name).value, self.p_bool_value)
        self.assertEqual(copied_instance.property(name=self.p_date_name).value, self.p_date_value)
        self.assertIn(
            self.p_datetime_value, copied_instance.property(name=self.p_datetime_name).value
        )
        self.assertEqual(copied_instance.property(name=self.p_link_name).value, self.p_link_value)
        self.assertEqual(
            copied_instance.property(name=self.p_single_select_name).value,
            self.p_single_select_value,
        )
        self.assertEqual(
            copied_instance.property(name=self.p_multi_select_name).value,
            self.p_multi_select_value,
        )
        list_of_user_ids = [
            user.id for user in copied_instance.property(name=self.p_user_reference_name).value
        ]
        self.assertIn(
            self.sub_part2.name,
            [
                part.name
                for part in copied_instance.property(name=self.p_part_reference_name).value
            ],
        )
        self.assertNotEqual(
            self.sub_part2.id,
            copied_instance.property(name=self.p_part_reference_name).value[0].id,
        )
        self.assertIn(
            self.front_wheel.id,
            [
                part.id
                for part in copied_instance.property(
                    name=self.p_part_reference_name_cross_scope
                ).value
            ],
        )
        self.assertDictEqual(
            self.p_geojson_value, copied_instance.property(name=self.p_geo_json_name).value
        )
        self.assertDictEqual(
            self.p_weather_value, copied_instance.property(name=self.p_weather_name).value
        )
        self.assertIn(
            self.specify_wheel_diameter.id,
            [
                act.id
                for act in copied_instance.property(name=self.p_activity_reference_name).value
            ],
        )
        self.assertIn(
            self.project.id,
            [
                scope.id
                for scope in copied_instance.property(name=self.p_project_reference_name).value
            ],
        )
        self.assertIn(self.test_manager.id, list_of_user_ids)
        self.assertIn(self.test_supervisor.id, list_of_user_ids)

        self.assertEqual(len(copied_instance._cached_children), 2)

    def test_copy_part_model_empty_name_not_include_children(self):
        # setUp
        model_target_parent = self.project.model("Bike")
        name_of_part = "__Copied model under Bike"
        copied_model = self.model_to_be_copied.copy(
            target_parent=model_target_parent,
            name=name_of_part,
            include_children=False,
            include_instances=False,
        )

        copied_model.populate_descendants()
        self.dump_part = copied_model

        # testing
        self.assertTrue(copied_model)
        self.assertEqual(copied_model.name, name_of_part)
        self.assertEqual(len(copied_model.properties), len(self.model_to_be_copied.properties))
        self.assertEqual(len(copied_model._cached_children), 0)

    def test_copy_internal_references_on_model(self):
        child_model = self.model_to_be_copied.children()[0]
        self.model_to_be_copied.add_property(
            name="__Property internal reference",
            default_value=child_model,
            property_type=PropertyType.REFERENCES_VALUE,
        )

        model_target_parent = self.project.model("Bike")
        self.dump_part = self.model_to_be_copied.copy(
            target_parent=model_target_parent,
            name="__Copied model under Bike",
            include_children=True,
            include_instances=False,
        )

        self.dump_part.refresh()
        copied_child = self.dump_part.children()[0]
        reference_property = self.dump_part.property(name="__Property internal reference")

        self.assertEqual(copied_child, reference_property.value[0])

    def test_copy_internal_references_on_instance(self):
        prop_name = "__Property internal reference"
        child_model = self.model_to_be_copied.children()[0]
        self.model_to_be_copied.add_property(
            name=prop_name, default_value=child_model, property_type=PropertyType.REFERENCES_VALUE
        )

        self.instance_to_be_copied.refresh()  # to load the new property
        self.instance_to_be_copied.property(
            name=prop_name
        ).value = self.instance_to_be_copied.children()[0]

        instance_target_parent = self.project.part("Bike")
        copied_instance = self.instance_to_be_copied.copy(
            target_parent=instance_target_parent,
            name="__Copied instance under Bike",
            include_children=True,
        )
        self.dump_part = copied_instance.model()

        copied_child = copied_instance.children()[0]
        reference_property = copied_instance.property(name=prop_name)

        self.assertTrue(reference_property.has_value())
        self.assertEqual(copied_child, reference_property.value[0])

    def test_move_part_model(self):
        # setUp
        model_target_parent = self.project.model("Bike")
        original_model_to_be_copied_name = self.model_to_be_copied.name
        self.model_to_be_copied = self.model_to_be_copied.move(
            target_parent=model_target_parent, name="__New part under bike", include_children=True
        )

        # testing
        with self.assertRaises(NotFoundError):
            self.project.model(name=original_model_to_be_copied_name)

    def test_copy_part_instance(self):
        # setUp
        instance_to_be_copied = self.model_to_be_copied.instances()[0]
        instance_to_be_copied.property(name=self.p_attach_name).upload(self.p_attach_value)
        instance_to_be_copied.property(name=self.p_part_reference_name_cross_scope).value = [
            self.front_wheel,
            self.rear_wheel,
        ]

        instance_target_parent = self.project.part("Bike")
        copied_instance = instance_to_be_copied.copy(
            target_parent=instance_target_parent, name="__Copied instance", include_children=True
        )
        copied_instance.populate_descendants()
        self.dump_part = copied_instance.model()

        # testing
        self.assertTrue(copied_instance)
        self.assertEqual(copied_instance.name, "__Copied instance")
        self.assertEqual(len(copied_instance.properties), len(self.model_to_be_copied.properties))
        self.assertEqual(copied_instance.property(name=self.p_char_name).value, self.p_char_value)
        self.assertEqual(
            copied_instance.property(name=self.p_multi_name).value, self.p_multi_value
        )
        self.assertEqual(copied_instance.property(name=self.p_int_name).value, self.p_int_value)
        self.assertEqual(
            copied_instance.property(name=self.p_float_name).value, self.p_float_value
        )
        self.assertEqual(copied_instance.property(name=self.p_bool_name).value, self.p_bool_value)
        self.assertEqual(copied_instance.property(name=self.p_date_name).value, self.p_date_value)
        self.assertIn(
            self.p_datetime_value, copied_instance.property(name=self.p_datetime_name).value
        )
        self.assertIn('requirements.txt',
                      copied_instance.property(name=self.p_attach_name).filename)
        self.assertIn('requirements.txt', self.p_attach_value)
        self.assertEqual(copied_instance.property(name=self.p_link_name).value, self.p_link_value)
        self.assertEqual(
            copied_instance.property(name=self.p_single_select_name).value,
            self.p_single_select_value,
        )
        self.assertEqual(
            copied_instance.property(name=self.p_multi_select_name).value,
            self.p_multi_select_value,
        )
        list_of_parts_ids = [
            part.id
            for part in copied_instance.property(name=self.p_part_reference_name_cross_scope).value
        ]
        list_of_user_ids = [
            user.id for user in copied_instance.property(name=self.p_user_reference_name).value
        ]
        self.assertIn(self.front_wheel.id, list_of_parts_ids)
        self.assertIn(self.rear_wheel.id, list_of_parts_ids)
        self.assertIn(
            self.specify_wheel_diameter.id,
            [
                act.id
                for act in copied_instance.property(name=self.p_activity_reference_name).value
            ],
        )
        self.assertIn(
            self.project.id,
            [
                scope.id
                for scope in copied_instance.property(name=self.p_project_reference_name).value
            ],
        )
        self.assertIn(self.test_manager.id, list_of_user_ids)
        self.assertIn(self.test_supervisor.id, list_of_user_ids)

        self.assertEqual(len(copied_instance._cached_children), 2)

    def test_move_part_instance(self):
        # setUp
        multiplicity_model = self.model_to_be_copied.add_model(
            name="__multiplicity part", multiplicity=Multiplicity.ZERO_MANY
        )
        multiplicity_part1 = self.instance_to_be_copied.add(
            model=multiplicity_model,
            name="__multiplicity part instance 1",
        )
        multiplicity_part2 = self.instance_to_be_copied.add(
            model=multiplicity_model,
            name="__multiplicity part instance 2",
        )

        clone_of_original_instance = multiplicity_part2.clone()

        moved_instance = clone_of_original_instance.move(
            target_parent=self.project.part("Bike"),
            name="__Moved clone instance",
            include_children=False,
        )

        moved_instance.populate_descendants()
        self.dump_part = moved_instance.model()

        # testing
        with self.assertRaises(NotFoundError):
            self.project.model(name=clone_of_original_instance.name)

        self.assertTrue(moved_instance)
        self.assertEqual(len(moved_instance.properties), 0)

    def test_copy_different_categories(self):
        target_instance = self.project.part(name="Bike")

        with self.assertRaisesRegex(IllegalArgumentError, "must have the same category"):
            self.model_to_be_copied.copy(target_parent=target_instance)

    def test_move_different_categories(self):
        target_instance = self.project.part(name="Bike")

        with self.assertRaisesRegex(IllegalArgumentError, "must have the same category"):
            self.model_to_be_copied.move(target_parent=target_instance)

    def test_copy_target_parent_inside_tree(self):
        with self.assertRaises(IllegalArgumentError):
            self.model_to_be_copied.copy(
                target_parent=self.sub_part1,
            )

    def test_copy_missing_target_parent_instance(self):
        target_parent = self.project.product_root_model.add_model(
            name="__ZERO+ MODEL",
            multiplicity=Multiplicity.ZERO_MANY,
        )
        self.dump_part = target_parent
        with self.assertRaises(IllegalArgumentError):
            self.model_to_be_copied.copy(
                target_parent=target_parent,
                include_instances=True,
            )

    def test_copy_too_many_target_parent_instances(self):
        target_parent = self.project.product_root_model.add_model(
            name="__ONE+ MODEL",
            multiplicity=Multiplicity.ONE_MANY,
        )
        self.project.product_root_instance.add(model=target_parent)
        self.dump_part = target_parent
        with self.assertRaises(IllegalArgumentError):
            self.model_to_be_copied.copy(
                target_parent=target_parent,
                include_instances=True,
            )

    def test_copy_attachments(self):
        names = [
            "__Property attachment_1",
            "__Property attachment_2",
        ]

        for name in names:
            self.model_to_be_copied.add_property(
                name=name,
                property_type=PropertyType.ATTACHMENT_VALUE,
            )

        self.instance_to_be_copied.refresh()

        for name in names:
            p_attachment = self.instance_to_be_copied.property(name)
            file = os.path.join(os.path.dirname(os.path.dirname(__file__)), "requirements.txt")
            p_attachment.upload(file)

        self.dump_part = self.model_to_be_copied.copy(
            target_parent=self.model_target_parent,
        )

        copied_instance = self.dump_part.instance()

        for name in names:
            copied_prop = copied_instance.property(name)
            self.assertTrue(copied_prop.has_value())

    def test_cross_scope_copy(self):
        name_of_part = "__Copied model under Bike"
        self.dump_part = self.model_to_be_copied.copy(
            target_parent=self.cross_scope_bike,
            name=name_of_part,
            include_children=True,
            include_instances=True,
        )

        self.dump_part.refresh()
        copied_options = self.dump_part.property(self.p_part_reference_name_cross_scope)._options
        self.assertTrue(self.model_to_be_copied.scope_id == copied_options.get("scope_id"))

        copied_sub_part2_int_prop = self.dump_part(self.sub_part2.name).property(self.p_int_name)
        copied_sub_part2_bool_prop = self.dump_part(self.sub_part2.name).property(self.p_bool_name)

        copied_part_ref_options = self.dump_part.property(self.p_part_reference_name)._options
        copied_scope_ref_options = self.dump_part.property(self.p_project_reference_name)._options

        self.assertTrue(self.dump_part.scope_id == copied_part_ref_options.get("scope_id"))
        self.assertTrue(
            copied_sub_part2_bool_prop.id in copied_part_ref_options.get("propmodels_excl"))
        self.assertTrue(
            copied_sub_part2_int_prop.id in copied_part_ref_options["prefilters"]["property_value"]
        )
        self.assertEqual(ScopeStatus.ACTIVE, copied_scope_ref_options["prefilters"]["status__in"])
        self.assertEqual("bike", copied_scope_ref_options["prefilters"]["tags__contains"])



from abc import abstractmethod, ABC
from typing import Optional, Any, Union, Tuple, List

from pykechain.exceptions import IllegalArgumentError
from pykechain.models import Property, Base
from pykechain.models.base import BaseInScope
from pykechain.models.input_checks import check_list_of_base, check_base
from pykechain.models.value_filter import BaseFilter


class _ReferenceProperty(Property):
    """
    Private base class for the KE-chain reference properties.

    .. versionadded:: 3.7
    :cvar REFERENCED_CLASS: handle to the Pykechain class that is referenced using this ReferenceProperty class.
    """

    REFERENCED_CLASS: type(Base) = Base

    def __init__(self, json, **kwargs):
        """Construct a ReferenceProperty from a json object."""
        super().__init__(json, **kwargs)

        self._cached_values = None

    @property
    def value(self) -> Optional[List[REFERENCED_CLASS]]:
        """
        Retrieve the referenced objects of this reference property.

        :return: list or generator of `Base` objects.
        :rtype list
        """
        if not self._value:
            return None
        elif not self._cached_values:
            self._cached_values = self._retrieve_objects()
        return self._cached_values

    @value.setter
    def value(self, value: Any) -> None:
        value = self.serialize_value(value)
        if self.use_bulk_update:
            self._pend_update(dict(value=value))
            self._value = [dict(id=pk) for pk in value] if isinstance(value, list) else None
            self._cached_values = None
        else:
            self._put_value(value)

    def value_ids(self) -> Optional[List[str]]:
        """
        Retrieve the referenced object UUIDs only.

        :return: list of UUIDs
        :rtype list
        """
        return [value.get("id") for value in self._value] if self.has_value() else None

    def _validate_values(self) -> List[str]:
        """
        Check if the `_value` attribute has valid content.

        :return list of UUIDs:
        :rtype list
        """
        if not self._value:
            return []

        object_ids = []
        for value in self._value:
            if isinstance(value, dict) and "id" in value:
                object_ids.append(str(value.get("id")))
            elif isinstance(value, (int, str)):
                object_ids.append(str(value))
            else:  # pragma: no cover
                raise ValueError(f'Value "{value}" must be a dict with field `id` or a UUID.')
        return object_ids

    @abstractmethod
    def _retrieve_objects(self, **kwargs) -> List[Base]:
        """
        Retrieve a list of Pykechain objects, type depending on the reference property type.

        This method is abstract because the exact retrieval method of the Client class changes per subclass.

        :param kwargs: optional arguments
        :return: list of Pykechain objects inheriting from Base
        """
        pass  # pragma: no cover

    def serialize_value(self, value: Union[Base, List, Tuple]) -> Optional[List[str]]:
        """
        Serialize the value to be set on the property by checking for a list of Base objects.

        :param value: non-serialized value
        :type value: Any
        :return: serialized value
        """
        try:
            value = check_base(value, cls=self.REFERENCED_CLASS, key="Reference")
            return [value] if value is not None else value
        except IllegalArgumentError:
            # expected to fail, as value should be an iterable
            return check_list_of_base(value, cls=self.REFERENCED_CLASS, key="references")

    def set_prefilters(
        self,
        prefilters: Optional[List[BaseFilter]] = None,
        clear: Optional[bool] = False,
    ):
        """
        Set the prefilters on the reference property.

        :return: None
        """
        raise NotImplementedError(f"Method not (yet) implemented for {self.__class__}")

    def get_prefilters(self):
        """
        Get the prefilters currently on the reference property.

        :return: list of filter values.
        """
        raise NotImplementedError(f"Method not (yet) implemented for {self.__class__}")


class _ReferencePropertyInScope(_ReferenceProperty, ABC):
    """
    Private base class for the KE-chain reference properties pointing to objects confined to a scope.

    .. versionadded:: 3.7
    """

    REFERENCED_CLASS = BaseInScope

    def _put_value(self, value: Union[List, Tuple]):
        super()._put_value(value=value)

        if value is not None:
            self._check_x_scope_id(referenced_object=value)

    def _check_x_scope_id(self, referenced_object: Union[Any, BaseInScope]) -> None:
        """
        Check whether this reference property has the `scope_id` in its `value_options`.

        :param referenced_object: Either a pykechain object or a UUID string.
        :return: None
        """
        if self._options and "scope_id" not in self._options:
            # See whether the referenced model is an object with an ID
            if isinstance(referenced_object, self.REFERENCED_CLASS):
                x_scope_id = referenced_object.scope_id
            else:
                # retrieve the scope_id from the property model's value (which is an object in a scope (x_scope))
                referenced_models = self.model().value
                if not referenced_models or not isinstance(
                    referenced_models[0], self.REFERENCED_CLASS
                ):
                    # if the referenced model is not set or the referenced value is not in current scope
                    x_scope_id = self.scope_id
                else:
                    # get the scope_id from the referenced model
                    x_scope_id = referenced_models[0].scope_id
            self._options["scope_id"] = x_scope_id

            # edit the model of the property, such that all instances are updated as well.
            self.model().edit(options=self._options)



"""Widgets for KE-chain 3."""
from .widget import Widget  # noqa: F403,F401,D104
from .widget_models import (
    MetapanelWidget,
    PropertygridWidget,
    UndefinedWidget,
    FilteredgridWidget,
    SupergridWidget,
    AttachmentviewerWidget,
    TasknavigationbarWidget,
    HtmlWidget,
    ServiceWidget,
    NotebookWidget,
    JsonWidget,
    MulticolumnWidget,
    ProgressWidget,
    ScopeWidget,
    ThirdpartyWidget,
    SignatureWidget,
    CardWidget,
    ServicecardWidget,
    DashboardWidget,
    TasksWidget,
    ScopemembersWidget,
)
from .widgets_manager import WidgetsManager

__all__ = (
    "Widget",
    "MetapanelWidget",
    "PropertygridWidget",
    "UndefinedWidget",
    "FilteredgridWidget",
    "SupergridWidget",
    "AttachmentviewerWidget",
    "TasknavigationbarWidget",
    "HtmlWidget",
    "ServiceWidget",
    "NotebookWidget",
    "JsonWidget",
    "MulticolumnWidget",
    "ProgressWidget",
    "SignatureWidget",
    "ScopeWidget",
    "CardWidget",
    "ServicecardWidget",
    "DashboardWidget",
    "TasksWidget",
    "ScopemembersWidget",
    "ThirdpartyWidget",
    "WidgetsManager",
)



import datetime
import time
import warnings
from test.support import EnvironmentVarGuard
from unittest import TestCase

import pytz

from pykechain.client import Client
from pykechain.enums import ScopeStatus
from pykechain.exceptions import (
    APIError,
    ClientError,
    ForbiddenError,
    IllegalArgumentError,
    NotFoundError,
)
from pykechain.models import Base, Team
from pykechain.models.scope import Scope
from tests.classes import TestBetamax


class TestClient(TestCase):
    def setUp(self):
        self.env = EnvironmentVarGuard()

    def test_init_default_url(self):
        client = Client()

        self.assertEqual(client.api_root, "http://localhost:8000/")

    def test_init_custom_url(self):
        client = Client(url="http://testing.com:1234")

        self.assertEqual(client.api_root, "http://testing.com:1234")

    def test_init_no_login(self):
        client = Client()

        self.assertNotIn("Authorization", client.headers)
        self.assertIsNone(client.auth)

    def test_init_basic_auth(self):
        client = Client()

        client.login("someuser", "withpass")

        self.assertNotIn("Authorization", client.headers)
        self.assertTrue(client.auth)

    def test_init_token(self):
        client = Client()
        PSEUDO_TOKEN = "123123"

        client.login(token=PSEUDO_TOKEN)

        self.assertTrue(client.headers["Authorization"], f"Token {PSEUDO_TOKEN}")
        self.assertIsNone(client.auth)

    def test_init_no_ssl(self):
        client = Client(check_certificates=False)

        self.assertFalse(client.session.verify)

    # 1.12
    def test_client_raises_error_with_false_url(self):
        with self.assertRaises(ClientError):
            Client(url="wrongurl")

    def test_client_from_env(self):
        """
        Test if the client does not provide a warning when no env file can be found

            Userwarning: 'Could not any envfile.'
            '.../lib/python3.5/site-packages/envparse.py'
        """
        self.env.set("KECHAIN_URL", "http://localhost:8000")
        with self.env:
            with warnings.catch_warnings(record=True) as captured_warnings:
                client = Client.from_env()
                self.assertEqual(len(captured_warnings), 0)

    # noinspection PyTypeChecker
    def test_reload(self):
        client = Client()

        not_a_kechain_object = 3
        with self.assertRaises(
            IllegalArgumentError, msg="Reload must receive an object of type Base."
        ):
            client.reload(not_a_kechain_object)

        empty_kechain_object = Base(json=dict(name="empty", id="1234567890"), client=client)
        with self.assertRaises(
            IllegalArgumentError, msg="Reload cant find API resource for type Base"
        ):
            client.reload(empty_kechain_object)


class TestClientLive(TestBetamax):
    def setUp(self):
        super().setUp()
        self.temp_scope = None

    def tearDown(self):
        """Teardown the test, remove the scope.

        When you create a scope, assign it to self.temp_scope such that it is deleted when the test is complete
        """
        if self.temp_scope:
            self.temp_scope.delete()
        super().tearDown()

    def test_login(self):
        self.assertTrue(self.project.parts())

    def test_no_login(self):
        self.client.login("wrong", "user")

        with self.assertRaises(ForbiddenError):
            self.client.parts()

    # 3.6.3
    def test_get_current_user(self):
        user = self.client.current_user()

        from pykechain.models import User

        self.assertIsInstance(user, User)

        self.client.auth = None
        self.client.headers = None
        with self.assertRaises(APIError):
            self.client.current_user()

    # 3.7.0
    def test_reload_deleted_object(self):
        bike_model = self.project.model(name="Bike")
        bike = bike_model.instance()
        wheel_model = bike_model.child(name="Wheel")

        wheel_name = "__new wheel: test for reloading"
        new_wheel_1 = bike.add(name=wheel_name, model=wheel_model)
        new_wheel_2 = self.project.part(name=wheel_name)

        self.assertFalse(new_wheel_1 is new_wheel_2, "Parts must be separate in Python memory")
        self.assertEqual(new_wheel_1, new_wheel_2, "Parts must be identical in UUIDs")

        new_wheel_2.delete()
        with self.assertRaises(NotFoundError):
            self.project.part(name=wheel_name)

        with self.assertRaises(NotFoundError, msg="Wheel cant be reloaded after it was deleted"):
            self.client.reload(new_wheel_1)

    # 2.6.0
    def test_create_scope(self):
        # setUp
        client = self.client
        scope_name = "New scope (pykechain testing - remove me)"
        scope_description = "This is a new scope for testing"
        scope_status = ScopeStatus.ACTIVE
        scope_tags = ["test_tag", "new_project_tag"]
        scope_start_date = datetime.datetime(2019, 4, 12, tzinfo=pytz.UTC)
        scope_due_date = datetime.datetime(2020, 4, 12, tzinfo=pytz.UTC)
        scope_team = client.teams()[0]

        self.temp_scope = client.create_scope(
            name=scope_name,
            description=scope_description,
            status=scope_status,
            tags=scope_tags,
            start_date=scope_start_date,
            due_date=scope_due_date,
            team=scope_team,
        )

        # testing
        self.assertEqual(self.temp_scope.name, scope_name)
        self.assertEqual(scope_start_date, self.temp_scope.start_date)
        self.assertEqual(scope_due_date, self.temp_scope.due_date)
        self.assertIn(scope_tags[0], self.temp_scope.tags)
        self.assertIn(scope_tags[1], self.temp_scope.tags)
        self.assertEqual(scope_team, self.temp_scope.team)

    def test_create_scope_with_team_name(self):
        # setUp
        team_name = self.client.teams()[0].name

        self.temp_scope = self.client.create_scope(
            name="New scope using the team name", team=team_name
        )

        # testing
        self.assertTrue(self.temp_scope.team)
        self.assertIsInstance(self.temp_scope.team, Team)
        self.assertEqual(team_name, self.temp_scope.team.name)

    def test_create_scope_with_team_uuid(self):
        # setUp
        team_id = self.client.teams()[0].id

        self.temp_scope = self.client.create_scope(
            name="New scope using the team ID", team=team_id
        )

        # testing
        self.assertTrue(self.temp_scope.team)
        self.assertIsInstance(self.temp_scope.team, Team)
        self.assertEqual(team_id, self.temp_scope.team.id)

    def test_create_scope_no_arguments(self):
        # setUp
        self.temp_scope = self.client.create_scope(name="New scope no arguments")

        # testing
        self.assertEqual(self.temp_scope.name, "New scope no arguments")
        self.assertTrue(self.temp_scope._json_data["start_date"])
        self.assertFalse(self.temp_scope.tags)

    # noinspection PyTypeChecker
    def test_create_scope_with_wrong_arguments(self):
        # testing
        with self.assertRaises(IllegalArgumentError):
            self.client.create_scope(name=12)
        with self.assertRaises(IllegalArgumentError):
            self.client.create_scope(name="Failed scope", status="LIMBO")
        with self.assertRaises(IllegalArgumentError):
            self.client.create_scope(name="Failed scope", description=True)
        with self.assertRaises(IllegalArgumentError):
            self.client.create_scope(name="Failed scope", tags="One tag no list")
        with self.assertRaises(IllegalArgumentError):
            self.client.create_scope(name="Failed scope", tags=[12, "this", "fails"])
        with self.assertRaises(IllegalArgumentError):
            self.client.create_scope(name="Failed scope", team=["Fake team"])

    def test_clone_scope(self):
        # setUp
        self.temp_scope = self.client.clone_scope(source_scope=self.project)

        # testing
        self.assertIsInstance(self.temp_scope, Scope)
        self.assertEqual(self.temp_scope.name, f"CLONE - {self.project.name}")
        self.assertEqual(self.temp_scope.status, self.project.status)
        self.assertEqual(self.temp_scope.start_date, self.project.start_date)
        self.assertEqual(self.temp_scope.due_date, self.project.due_date)
        self.assertEqual(self.temp_scope.description, self.project.description)
        self.assertListEqual(self.temp_scope.tags, self.project.tags)

    def test_clone_scope_with_arguments(self):
        # setUp
        now = datetime.datetime.now()
        name = "test clone"
        status = ScopeStatus.CLOSED
        description = "test description"
        tags = ["one", "two"]
        team = self.project.team

        self.temp_scope = self.client.clone_scope(
            name=name,
            source_scope=self.project,
            status=status,
            start_date=now,
            due_date=now,
            description=description,
            tags=tags,
            team=team,
        )

        self.assertIsInstance(self.temp_scope, Scope)
        self.assertEqual(self.temp_scope.name, name)
        self.assertEqual(self.temp_scope.status, status)
        self.assertEqual(self.temp_scope.start_date, now)
        self.assertEqual(self.temp_scope.due_date, now)
        self.assertEqual(self.temp_scope.description, description)
        self.assertEqual(self.temp_scope.team, team)
        self.assertListEqual(self.temp_scope.tags, tags)

    def test_scope_delete(self):
        new_scope = self.project.clone(asynchronous=False)
        self.assertNotEqual(self.project.id, new_scope.id)
        new_scope.delete(asynchronous=False)

        with self.assertRaisesRegex(NotFoundError, "fit criteria"):
            # throw in arbitrary sleep to give backend time to actually delete the scope.
            time.sleep(1)
            self.client.scope(pk=new_scope.id)


class TestCloneScopeAsync(TestBetamax):
    def setUp(self):
        super().setUp()
        self.temp_scope = None
        self.source = self.client.create_scope(
            name="_Async to be cloned scope SOURCE",
        )
        time.sleep(3)

    def tearDown(self):
        self.source.delete()
        if self.temp_scope:
            self.temp_scope.delete()
        super().tearDown()

    def test_clone_asynchronous(self):
        # setUp
        clone_name = "_Async cloned scope TARGET"

        self.client.clone_scope(
            name=clone_name,
            source_scope=self.source,
            asynchronous=True,
        )

        for _ in range(5):
            try:
                self.temp_scope = self.client.scope(name=clone_name)
            except NotFoundError:
                time.sleep(3)
                continue
            else:
                break

        # testing
        self.assertIsInstance(self.temp_scope, Scope)


class TestClientAppVersions(TestBetamax):
    def test_retrieve_versions(self):
        """Test to retrieve the app versions from KE-chain"""
        app_versions = self.project._client.app_versions
        self.assertTrue(isinstance(app_versions, list))
        self.assertTrue(isinstance(app_versions[0], dict))
        self.assertTrue(
            set(app_versions[0].keys()),
            {"app", "label", "version", "major", "minor", "patch", "prerelease"},
        )

    def test_compare_versions(self):
        """Multitest to check all the matchings versions"""

        self.assertTrue(self.client.match_app_version(app="kechain2.core.wim", version=">=1.0.0"))
        self.assertTrue(self.client.match_app_version(label="wim", version=">=1.0.0"))
        self.assertFalse(self.client.match_app_version(label="wim", version="==0.0.1"))

        # value error
        # wrong version string (no semver string) to check against
        with self.assertRaises(ValueError):
            self.client.match_app_version(app="kechain2.core.wim", version="1.0")

        # right version, no operand in version
        with self.assertRaises(ValueError):
            self.client.match_app_version(app="kechain2.core.wim", version="1.0.0")

        # wrong operand (should be ==)
        with self.assertRaises(ValueError):
            self.client.match_app_version(app="kechain2.core.wim", version="=1.0.0")

        # not found a match version
        with self.assertRaises(IllegalArgumentError):
            self.client.match_app_version(app="kechain2.core.wim", version="")

        # no version found on the app kechain2.metrics
        with self.assertRaises(NotFoundError):
            self.client.match_app_version(
                app="kechain2.core.activitylog", version=">0.0.0", default=None
            )

        # no version found on the app kechain2.metrics, default = True, returns True
        self.assertTrue(
            self.client.match_app_version(
                app="kechain2.core.activitylog", version=">0.0.0", default=True
            )
        )

        # no version found on the app kechain2.metrics, default = False, returns False
        self.assertFalse(
            self.client.match_app_version(
                app="kechain2.core.activitylog", version=">0.0.0", default=False
            )
        )

        # no version found on the app kechain2.metrics, default = False, returns False
        # default is set to return False in the method
        self.assertFalse(
            self.client.match_app_version(app="kechain2.core.activitylog", version=">0.0.0")
        )

        # did not find the app
        with self.assertRaises(NotFoundError):
            self.client.match_app_version(app="nonexistingapp", version=">0.0.0", default=None)

        # did not find the app, but the default returns a False
        self.assertFalse(
            self.client.match_app_version(app="nonexistingapp", version=">0.0.0", default=False)
        )

        # did not find the app, but the default returns a False, without providing the default, as the default is False
        self.assertFalse(self.client.match_app_version(app="nonexistingapp", version=">0.0.0"))



from unittest import TestCase


class TestAbout(TestCase):
    def test_import(self):
        from pykechain import __about__ as about

        self.assertTrue(about)

    def test_names(self):
        from pykechain import __about__ as about

        self.assertTrue(about.name)
        self.assertTrue(about.version)
        import semver

        self.assertTrue(semver.VersionInfo.parse(about.version))

    def test_python_fstring(self):
        true = f"this is {True}"
        self.assertEqual(f"{true}", true)



import datetime
import os
from unittest import TestCase

import pytz
import six
from betamax import Betamax

if six.PY2:
    from test.test_support import EnvironmentVarGuard
else:
    from test.support import EnvironmentVarGuard

from pykechain import Client
from tests.utils import TEST_TOKEN, TEST_SCOPE_NAME, TEST_URL, TEST_RECORD_CASSETTES, TEST_SCOPE_ID

with Betamax.configure() as config:
    config.cassette_library_dir = os.path.join(os.path.dirname(__file__), "cassettes")
    config.define_cassette_placeholder("<API_URL>", TEST_URL)
    config.define_cassette_placeholder("<AUTH_TOKEN>", TEST_TOKEN)


class TestBetamax(TestCase):
    time = datetime.datetime(year=2020, month=1, day=1, tzinfo=pytz.UTC)

    @property
    def cassette_name(self):
        test = self._testMethodName
        return f"{self.__class__.__name__}.{test}"

    def setUp(self):
        # use self.env.set('var', 'value') and with self.env: ... to use custom environmental variables
        self.env = EnvironmentVarGuard()
        self.client = Client(url=TEST_URL)

        if TEST_TOKEN:
            self.client.login(token=TEST_TOKEN)

        self.recorder = Betamax(session=self.client.session)

        if TEST_RECORD_CASSETTES:
            self.recorder.use_cassette(self.cassette_name)
            self.recorder.start()
        else:
            self.recorder.config.record_mode = None

        if TEST_SCOPE_NAME:
            self.project = self.client.scope(name=TEST_SCOPE_NAME)
        elif TEST_SCOPE_ID:
            self.project = self.client.scope(id=TEST_SCOPE_ID)
        else:
            raise Exception(
                "Could not retrieve the test scope, you need to provide a "
                "`TEST_SCOPE_ID` or `TEST_SCOPE_NAME` in your `.env` file"
            )

    def tearDown(self):
        self.recorder.stop()
        del self.env
        del self.client
